# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Convolution layers."""

from typing import Optional, Tuple, Union, List

from torch import Tensor, arange, cat, float64, int32, ones
from torch.nn.functional import pad, unfold
from torch.nn.modules.conv import _ConvNd, Conv1d, Conv2d, Conv3d
from torch.nn.modules.utils import _single, _pair, _triple

from aihwkit.nn.functions import AnalogIndexedFunction, AnalogFunction
from aihwkit.nn.modules.base import AnalogModuleBase, RPUConfigAlias
from aihwkit.simulator.configs import SingleRPUConfig


class _AnalogConvNd(AnalogModuleBase, _ConvNd):
    """Base class for convolution layers."""

    __constants__ = ['stride', 'padding', 'dilation', 'groups',
                     'padding_mode', 'output_padding', 'in_channels',
                     'out_channels', 'kernel_size', 'in_features', 'out_features',
                     'realistic_read_write',
                     'digital_bias', 'analog_bias', 'use_bias']
    in_channels: int
    out_channels: int
    kernel_size: Tuple[int, ...]
    stride: Tuple[int, ...]
    padding: Tuple[int, ...]
    dilation: Tuple[int, ...]
    realistic_read_write: bool
    transposed: bool
    output_padding: Tuple[int, ...]
    groups: int
    padding_mode: str
    fold_indices: Tensor
    input_size: float
    in_features: int
    out_features: int
    digital_bias: bool
    analog_bias: bool
    use_bias: bool

    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            kernel_size: Tuple[int, ...],
            stride: Tuple[int, ...],
            padding: Tuple[int, ...],
            dilation: Tuple[int, ...],
            transposed: bool,
            output_padding: Tuple[int, ...],
            groups: int,
            bias: bool,
            padding_mode: str,
            rpu_config: Optional[RPUConfigAlias] = None,
            realistic_read_write: bool = False,
            weight_scaling_omega: Optional[bool] = None,
            use_indexed: Optional[bool] = None
    ):
        # pylint: disable=too-many-arguments, too-many-locals
        if groups != 1:
            raise ValueError('Only one group is supported')
        if padding_mode != 'zeros':
            raise ValueError('Only "zeros" padding mode is supported')

        # Call super() after tile creation, including ``reset_parameters``.
        _ConvNd.__init__(self, in_channels, out_channels, kernel_size, stride,
                         padding, dilation, transposed, output_padding, groups, bias,
                         padding_mode)

        # Create the tile and set the analog.
        if rpu_config is None:
            rpu_config = SingleRPUConfig()

        rpu_config = self._set_weight_scaling_omega(rpu_config, weight_scaling_omega)

        AnalogModuleBase.__init__(
            self,
            self.get_tile_size(in_channels, groups, kernel_size),
            out_channels,
            bias,
            realistic_read_write,
            rpu_config.mapping
        )
        self.analog_tile = self._setup_tile(rpu_config)

        # Register analog tile
        self.register_analog_tile(self.analog_tile)

        # Set weights from the reset_parameters
        self.set_weights(self.weight, self.bias)

        # Set the index matrices.
        self.use_indexed = use_indexed

        self.fold_indices = Tensor().detach()
        self.register_helper('fold_indices')
        self.input_size = 0
        self.register_helper('input_size')
        self.tensor_view = (-1,)  # type: Tuple[int, ...]

        # Unregister weight/bias as a parameter but keep it for syncs
        self.unregister_parameter('weight')
        if self.analog_bias:
            self.unregister_parameter('bias')

    def get_tile_size(
            self,
            in_channels: int,
            groups: int,
            kernel_size: Tuple[int, ...]
    ) -> int:
        """Calculate the tile size."""
        raise NotImplementedError

    def get_image_size(self, size: int, i: int) -> int:
        """Calculate the output image sizes."""
        nom = (size + 2 * self.padding[i] - self.dilation[i] * (self.kernel_size[i] - 1) - 1)
        return nom // self.stride[i] + 1

    def reset_parameters(self) -> None:
        """Reset the parameters (weight and bias)."""
        super().reset_parameters()
        if self.analog_tile_count():
            self.set_weights(self.weight, self.bias)

    def _recalculate_indexes(self, x_input: Tensor) -> None:
        """Calculate and set the indexes of the analog tile."""

        self.fold_indices, image_sizes, self.input_size = \
            self._calculate_indexes(x_input, self.in_channels)
        self.analog_tile.set_indexed(self.fold_indices, image_sizes)

    def _calculate_indexes(self, x_input: Tensor,
                           in_channels: int) -> Tuple[Tensor, List[int], int]:
        """Calculate and return the fold indexes and sizes.

        Args:
            x_input: input matrix
            in_channels: number of input channel

        Returns:
            fold_indices: indices for the analog tile
            image_sizes: image sizes for the analog tile
            input_size: size of the current input
        """
        raise NotImplementedError

    def _forward_indexed(self, x_input: Tensor) -> Tensor:
        """Compute the forward pass in indexed fashion. This is fast and
        memory-efficient indexed convolution (only for GPUs)"""

        input_size = x_input.numel() / x_input.size(0)
        if self.input_size != input_size or not self.analog_tile.is_indexed():
            self._recalculate_indexes(x_input)

        return AnalogIndexedFunction.apply(
            self.analog_tile.get_analog_ctx(), x_input,
            self.analog_tile.shared_weights, not self.training)

    def _forward_unfold(self, x_input: Tensor) -> Tensor:
        """Forward using explicit unfolding (more suitable for CPUs) """
        im_shape = x_input.shape
        x_input_ = unfold(x_input, kernel_size=self.kernel_size, dilation=self.dilation,
                          padding=self.padding, stride=self.stride).transpose(1, 2)

        out = AnalogFunction.apply(
            self.analog_tile.get_analog_ctx(), x_input_,
            self.analog_tile.shared_weights, not self.training).transpose(1, 2)

        out_size = (im_shape[2] + 2 * self.padding[0]
                    - self.dilation[0] * (self.kernel_size[0] - 1) - 1) // self.stride[0] + 1
        return out.view(im_shape[0], self.out_channels, out_size, -1)

    def forward(self, x_input: Tensor) -> Tensor:
        """Compute the forward pass."""

        if self.use_indexed is None:
            use_indexed = self.analog_tile.device.type == 'cuda'
        else:
            use_indexed = self.use_indexed

        if use_indexed:
            out = self._forward_indexed(x_input)
        else:
            out = self._forward_unfold(x_input)

        out = self.analog_tile.apply_out_scaling(out, self.tensor_view)

        if self.digital_bias:
            return out + self.bias.view(*self.tensor_view)
        return out


class AnalogConv1d(_AnalogConvNd):
    """1D convolution layer that uses an analog tile.

    Applies a 1D convolution over an input signal composed of several input
    planes, using an analog tile for its forward, backward and update passes.

    Note:
        The tensor parameters of this layer (``.weight`` and ``.bias``) are not
        guaranteed to contain the same values as the internal weights and biases
        stored in the analog tile. Please use ``set_weights`` and
        ``get_weights`` when attempting to read or modify the weight/bias. This
        read/write process can simulate the (noisy and inexact) analog writing
        and reading of the resistive elements.

    Args:
        in_channels: number of channels in the input image.
        out_channels: number of channels produced by the convolution.
        kernel_size: size of the convolving kernel.
        stride: stride of the convolution.
        padding: zero-padding added to both sides of the input.
        dilation: spacing between kernel elements.
        groups: number of blocked connections from input channels to output
            channels.
        bias: whether to use a bias row on the analog tile or not.
        padding_mode: padding strategy. Only ``'zeros'`` is supported.
        rpu_config: resistive processing unit configuration.
        realistic_read_write: whether to enable realistic read/write
            for setting initial weights and during reading of the weights.
        weight_scaling_omega: depreciated, use
            :class:`aihwkit.simulator.configs.utils.MappingParameter`
            instead to specify weight scaling
    """
    # pylint: disable=abstract-method

    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            kernel_size: Union[int, Tuple],
            stride: Union[int, Tuple] = 1,
            padding: Union[int, Tuple] = 0,
            dilation: Union[int, Tuple] = 1,
            groups: int = 1,
            bias: bool = True,
            padding_mode: str = 'zeros',
            rpu_config: Optional[RPUConfigAlias] = None,
            realistic_read_write: bool = False,
            weight_scaling_omega: Optional[bool] = None
    ):
        # pylint: disable=too-many-arguments
        kernel_size = _single(kernel_size)
        stride = _single(stride)
        padding = _single(padding)
        dilation = _single(dilation)

        if dilation != _single(1):
            raise ValueError('Only dilation = 1 is supported')

        super().__init__(
            in_channels, out_channels, kernel_size, stride, padding, dilation,  # type: ignore
            False, _single(0), groups, bias, padding_mode,
            rpu_config, realistic_read_write, weight_scaling_omega, True
        )

        self.tensor_view = (-1, 1)

    @classmethod
    def from_digital(
            cls,
            module: Conv1d,
            rpu_config: Optional[RPUConfigAlias] = None,
            realistic_read_write: bool = False,
    ) -> 'AnalogConv1d':
        """Return an AnalogConv1d layer from a torch Conv1d layer.

        Args:
            module: The torch module to convert. All layers that are
                defined in the ``conversion_map``.
            rpu_config: RPU config to apply to all converted tiles.
                Applied to all converted tiles.
            realistic_read_write: Whether to use closed-loop programming
                when setting the weights. Applied to all converted tiles.

                Note:
                    Make sure that the weight max and min settings of the
                    device support the desired analog weight range.

        Returns:
            an AnalogConv1d layer based on the digital Conv1d ``module``.
        """
        analog_module = cls(module.in_channels,
                            module.out_channels,
                            module.kernel_size,
                            module.stride,
                            module.padding,
                            module.dilation,
                            module.groups,
                            module.bias is not None,
                            module.padding_mode,
                            rpu_config,
                            realistic_read_write,
                            )

        analog_module.set_weights(module.weight, module.bias)
        return analog_module

    def get_tile_size(
            self,
            in_channels: int,
            groups: int,
            kernel_size: Tuple[int, ...]
    ) -> int:
        """Calculate the tile size."""
        return (in_channels // groups) * kernel_size[0]

    def _calculate_indexes(self, x_input: Tensor,
                           in_channels: int) -> Tuple[Tensor, List[int], int]:
        """Calculate and return the fold indexes and sizes.

        Args:
            x_input: input matrix
            in_channels: number of input channel

        Returns:
            fold_indices: indices for the analog tile
            image_sizes: image sizes for the analog tile
            input_size: size of the current input
        """
        input_size = x_input.numel() / x_input.size(0)

        # pytorch just always uses NCHW order
        fold_indices = arange(2, x_input.size(2) + 2, dtype=float64).detach()
        shape = [1] + [1] + list(x_input.shape[2:])
        fold_indices = fold_indices.reshape(*shape)
        if not all(item == 0 for item in self.padding):
            fold_indices = pad(fold_indices, pad=[self.padding[0], self.padding[0]],
                               mode='constant', value=0)
        unfolded = fold_indices.unfold(2, self.kernel_size[0], self.stride[0]).clone()
        fold_indices = unfolded.reshape(-1, self.kernel_size[0]).transpose(0, 1).flatten().round()

        # concatenate the matrix index for different channels
        fold_indices_orig = fold_indices.clone()
        for i in range(in_channels - 1):
            fold_indices_tmp = fold_indices_orig.clone()
            for j in range(fold_indices_orig.size(0)):
                if fold_indices_orig[j] != 0:
                    fold_indices_tmp[j] += (input_size / in_channels) * (i + 1)

            fold_indices = cat([fold_indices, fold_indices_tmp], dim=0).clone()

        fold_indices = fold_indices.to(dtype=int32)

        if self.analog_bias:
            out_image_size = fold_indices.numel() // (self.kernel_size[0])
            fold_indices = cat((fold_indices, ones(out_image_size, dtype=int32)), 0)

        fold_indices = fold_indices.to(x_input.device)

        x_height = x_input.size(2)
        d_height = self.get_image_size(x_height, 0)

        image_sizes = [in_channels, x_height, d_height]
        return (fold_indices, image_sizes, input_size)


class AnalogConv2d(_AnalogConvNd):
    """2D convolution layer that uses an analog tile.

    Applies a 2D convolution over an input signal composed of several input
    planes, using an analog tile for its forward, backward and update passes.

    Note:
        The tensor parameters of this layer (``.weight`` and ``.bias``) are not
        guaranteed to contain the same values as the internal weights and biases
        stored in the analog tile. Please use ``set_weights`` and
        ``get_weights`` when attempting to read or modify the weight/bias. This
        read/write process can simulate the (noisy and inexact) analog writing
        and reading of the resistive elements.

    Args:
        in_channels: number of channels in the input image.
        out_channels: number of channels produced by the convolution.
        kernel_size: size of the convolving kernel.
        stride: stride of the convolution.
        padding: zero-padding added to both sides of the input.
        dilation: spacing between kernel elements.
        groups: number of blocked connections from input channels to output
            channels.
        bias: whether to use a bias row on the analog tile or not.
        padding_mode: padding strategy. Only ``'zeros'`` is supported.
        rpu_config: resistive processing unit configuration.
        realistic_read_write: whether to enable realistic read/write
            for setting initial weights and read out of weights.
        weight_scaling_omega: depreciated, use
            :class:`aihwkit.simulator.configs.utils.MappingParameter`
            instead to specify weight scaling
        use_indexed: Whether to use explicit unfolding or implicit indexing. If
            None (default), it will use implicit indexing for CUDA and
            explicit unfolding for CPU
    """
    # pylint: disable=abstract-method

    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            kernel_size: Union[int, Tuple],
            stride: Union[int, Tuple] = 1,
            padding: Union[int, Tuple] = 0,
            dilation: Union[int, Tuple] = 1,
            groups: int = 1,
            bias: bool = True,
            padding_mode: str = 'zeros',
            rpu_config: Optional[RPUConfigAlias] = None,
            realistic_read_write: bool = False,
            weight_scaling_omega: Optional[bool] = None,
            use_indexed: Optional[bool] = None
    ):
        # pylint: disable=too-many-arguments
        kernel_size = _pair(kernel_size)
        stride = _pair(stride)
        padding = _pair(padding)
        dilation = _pair(dilation)

        super().__init__(
            in_channels, out_channels, kernel_size, stride, padding, dilation,  # type: ignore
            False, _pair(0), groups, bias, padding_mode,
            rpu_config, realistic_read_write, weight_scaling_omega, use_indexed
        )

        self.tensor_view = (-1, 1, 1)

    @classmethod
    def from_digital(
            cls,
            module: Conv2d,
            rpu_config: Optional[RPUConfigAlias] = None,
            realistic_read_write: bool = False,
    ) -> 'AnalogConv2d':
        """Return an AnalogConv2d layer from a torch Conv2d layer.

        Args:
            module: The torch module to convert. All layers that are
                defined in the ``conversion_map``.
            rpu_config: RPU config to apply to all converted tiles.
                Applied to all converted tiles.
            realistic_read_write: Whether to use closed-loop programming
                when setting the weights. Applied to all converted tiles.

                Note:
                    Make sure that the weight max and min settings of the
                    device support the desired analog weight range.

        Returns:
            an AnalogConv2d layer based on the digital Conv2d ``module``.
        """
        analog_module = cls(module.in_channels,
                            module.out_channels,
                            module.kernel_size,
                            module.stride,
                            module.padding,
                            module.dilation,
                            module.groups,
                            module.bias is not None,
                            module.padding_mode,
                            rpu_config,
                            realistic_read_write,
                            )

        analog_module.set_weights(module.weight, module.bias)
        return analog_module

    def get_tile_size(
            self,
            in_channels: int,
            groups: int,
            kernel_size: Tuple[int, ...]
    ) -> int:
        """Calculate the tile size."""
        return (in_channels // groups) * kernel_size[0] * kernel_size[1]

    def _calculate_indexes(self, x_input: Tensor,
                           in_channels: int) -> Tuple[Tensor, List[int], int]:
        """Calculate and return the fold indexes and sizes.

        Args:
            x_input: input matrix
            in_channels: number of input channel

        Returns:
            fold_indices: indices for the analog tile
            image_sizes: image sizes for the analog tile
            input_size: size of the current input
        """
        input_size = x_input.numel() / x_input.size(0)

        # pytorch just always uses NCHW order
        fold_indices = arange(2, input_size + 2, dtype=float64).detach()
        shape = [1] + list(x_input.shape[1:])
        fold_indices = fold_indices.reshape(*shape)
        fold_indices = unfold(fold_indices,
                              kernel_size=self.kernel_size,
                              stride=self.stride,
                              padding=self.padding,
                              dilation=self.dilation).flatten().round().to(dtype=int32)

        if self.analog_bias:
            out_image_size = fold_indices.numel() // (self.kernel_size[0] * self.kernel_size[1])
            fold_indices = cat((fold_indices, ones(out_image_size, dtype=int32)), 0)

        fold_indices = fold_indices.to(x_input.device)

        x_height = x_input.size(2)
        x_width = x_input.size(3)

        d_height = self.get_image_size(x_height, 0)
        d_width = self.get_image_size(x_width, 1)

        image_sizes = [in_channels, x_height, x_width, d_height, d_width]
        return (fold_indices, image_sizes, input_size)


class AnalogConv3d(_AnalogConvNd):
    """3D convolution layer that uses an analog tile.

    Applies a 3D convolution over an input signal composed of several input
    planes, using an analog tile for its forward, backward and update passes.

    Note:
        The tensor parameters of this layer (``.weight`` and ``.bias``) are not
        guaranteed to contain the same values as the internal weights and biases
        stored in the analog tile. Please use ``set_weights`` and
        ``get_weights`` when attempting to read or modify the weight/bias. This
        read/write process can simulate the (noisy and inexact) analog writing
        and reading of the resistive elements.

    Args:
        in_channels: number of channels in the input image.
        out_channels: number of channels produced by the convolution.
        kernel_size: size of the convolving kernel.
        stride: stride of the convolution.
        padding: zero-padding added to both sides of the input.
        dilation: spacing between kernel elements.
        groups: number of blocked connections from input channels to output
            channels.
        bias: whether to use a bias row on the analog tile or not.
        padding_mode: padding strategy. Only ``'zeros'`` is supported.
        rpu_config: resistive processing unit configuration.
        realistic_read_write: whether to enable realistic read/write
            for setting initial weights and read out of weights.
        weight_scaling_omega: depreciated, use
            :class:`aihwkit.simulator.configs.utils.MappingParameter`
            instead to specify weight scaling
    """
    # pylint: disable=abstract-method

    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            kernel_size: Union[int, Tuple],
            stride: Union[int, Tuple] = 1,
            padding: Union[int, Tuple] = 0,
            dilation: Union[int, Tuple] = 1,
            groups: int = 1,
            bias: bool = True,
            padding_mode: str = 'zeros',
            rpu_config: Optional[RPUConfigAlias] = None,
            realistic_read_write: bool = False,
            weight_scaling_omega: Optional[bool] = None
    ):
        # pylint: disable=too-many-arguments
        kernel_size = _triple(kernel_size)
        stride = _triple(stride)
        padding = _triple(padding)
        dilation = _triple(dilation)

        if dilation != _triple(1):
            raise ValueError('Only dilation = 1 is supported')

        super().__init__(
            in_channels, out_channels, kernel_size, stride, padding, dilation,  # type: ignore
            False, _triple(0), groups, bias, padding_mode,
            rpu_config, realistic_read_write, weight_scaling_omega, True
        )

        self.tensor_view = (-1, 1, 1, 1)

    @classmethod
    def from_digital(
            cls,
            module: Conv3d,
            rpu_config: Optional[RPUConfigAlias] = None,
            realistic_read_write: bool = False,
    ) -> 'AnalogConv3d':
        """Return an AnalogConv3d layer from a torch Conv3d layer.

        Args:
            module: The torch module to convert. All layers that are
                defined in the ``conversion_map``.
            rpu_config: RPU config to apply to all converted tiles.
                Applied to all converted tiles.
            realistic_read_write: Whether to use closed-loop programming
                when setting the weights. Applied to all converted tiles.

                Note:
                    Make sure that the weight max and min settings of the
                    device support the desired analog weight range.

        Returns:
            an AnalogConv3d layer based on the digital Conv3d ``module``.
        """
        analog_module = cls(module.in_channels,
                            module.out_channels,
                            module.kernel_size,
                            module.stride,
                            module.padding,
                            module.dilation,
                            module.groups,
                            module.bias is not None,
                            module.padding_mode,
                            rpu_config,
                            realistic_read_write,
                            )

        analog_module.set_weights(module.weight, module.bias)
        return analog_module

    def get_tile_size(
            self,
            in_channels: int,
            groups: int,
            kernel_size: Tuple[int, ...]
    ) -> int:
        """Calculate the tile size."""
        return (in_channels // groups) * (
                kernel_size[0] * kernel_size[1] * kernel_size[2])

    def _calculate_indexes(self, x_input: Tensor,
                           in_channels: int) -> Tuple[Tensor, List[int], int]:
        """Calculate and return the fold indexes and sizes.

        Args:
            x_input: input matrix
            in_channels: then number of in channels

        Returns:
            fold_indices: indices for the analog tile
            image_sizes: image sizes for the analog tile
            input_size: size of the current input
        """
        # pylint: disable=too-many-locals
        input_size = x_input.numel() / x_input.size(0)

        # pytorch just always uses NCDHW order
        fold_indices = arange(2, x_input.size(2) * x_input.size(3) * x_input.size(4) + 2,
                              dtype=float64).detach()
        shape = [1] + [1] + list(x_input.shape[2:])
        fold_indices = fold_indices.reshape(*shape)
        if not all(item == 0 for item in self.padding):
            fold_indices = pad(fold_indices, pad=[
                self.padding[2], self.padding[2],
                self.padding[1], self.padding[1],
                self.padding[0], self.padding[0]], mode='constant', value=0)
        unfolded = fold_indices.unfold(2, self.kernel_size[0], self.stride[0]). \
            unfold(3, self.kernel_size[1], self.stride[1]). \
            unfold(4, self.kernel_size[2], self.stride[2]).clone()

        fold_indices = unfolded.reshape(-1, self.kernel_size[0] * self.kernel_size[1] *
                                        self.kernel_size[2]).transpose(0, 1).flatten().round()

        # concatenate the matrix index for different channels
        fold_indices_orig = fold_indices.clone()
        for i in range(in_channels - 1):
            fold_indices_tmp = fold_indices_orig.clone()
            for j in range(fold_indices_orig.size(0)):
                if fold_indices_orig[j] != 0:
                    fold_indices_tmp[j] += (input_size / in_channels) * (i + 1)

            fold_indices = cat([fold_indices, fold_indices_tmp], dim=0).clone()

        fold_indices = fold_indices.to(dtype=int32)

        if self.analog_bias:
            out_image_size = fold_indices.numel() // (self.kernel_size[0] *
                                                      self.kernel_size[1] *
                                                      self.kernel_size[2])
            fold_indices = cat((fold_indices, ones(out_image_size, dtype=int32)), 0)

        fold_indices = fold_indices.to(x_input.device)

        x_depth = x_input.size(2)
        x_height = x_input.size(3)
        x_width = x_input.size(4)

        d_depth = self.get_image_size(x_depth, 0)
        d_height = self.get_image_size(x_height, 1)
        d_width = self.get_image_size(x_width, 2)

        image_sizes = [in_channels, x_depth, x_height, x_width, d_depth, d_height, d_width]
        return (fold_indices, image_sizes, input_size)



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""RPU configurations presets used for the composer interface."""

# pylint: disable=too-many-instance-attributes

from dataclasses import dataclass, field
from aihwkit.simulator.configs.configs import (
    InferenceRPUConfig
)

from aihwkit.simulator.configs.utils import (
    IOParameters, WeightClipParameter,  WeightModifierParameter, MappingParameter,
    BoundManagementType, NoiseManagementType, WeightNoiseType, WeightModifierType,
    WeightClipType
)
from aihwkit.inference import (
    BaseDriftCompensation, BaseNoiseModel, GlobalDriftCompensation,
    PCMLikeNoiseModel
)


@dataclass
class WebComposerIOParameters(IOParameters):
    r"""Preset for the forward and backward pass parameters.

    The default values used for the web composer
    """

    bound_management: BoundManagementType = BoundManagementType.ITERATIVE
    noise_management: NoiseManagementType = NoiseManagementType.ABS_MAX

    inp_res: float = 1.0 / (2**7 - 2)  # 7 bit DAC.
    inp_sto_round: bool = False

    out_bound: float = 20.0
    out_noise: float = 0.1
    out_res: float = 1.0 / (2**9 - 2)  # 9 bit ADC.

    w_noise: float = 0.0
    w_noise_type: WeightNoiseType = WeightNoiseType.NONE


@dataclass
class WebComposerWeightModifierParameter(WeightModifierParameter):
    r"""Preset for the WeightModifierParameter

    The default values used for the web composer
    """

    type: WeightModifierType = WeightModifierType.ADD_NORMAL
    std_dev: float = 0.1


@dataclass
class WebComposerWeightClipParameter(WeightClipParameter):
    """Parameter that clip the weights during hardware-aware training.

    The default values used for the web composer.
    """

    fixed_value: float = 1.0
    type: WeightClipType = WeightClipType.FIXED_VALUE


@dataclass
class WebComposerMappingParameter(MappingParameter):
    """Parameter related to hardware design and the mapping of logical
    weight matrices to physical tiles.

    The default values used for the web composer.
    """

    digital_bias: bool = True

    weight_scaling_omega: float = 1.0
    weight_scaling_columnwise: bool = True
    learn_out_scaling: bool = True
    out_scaling_columnwise: bool = False

    max_input_size: int = 512
    max_output_size: int = 512


@dataclass
class WebComposerInferenceRPUConfig(InferenceRPUConfig):
    """Preset configuration used as default for the Inference Composer
    """

    forward: IOParameters = field(default_factory=WebComposerIOParameters)
    """Input-output parameter setting for the forward direction."""

    noise_model: BaseNoiseModel = field(default_factory=PCMLikeNoiseModel)
    """Statistical noise model to be used during (realistic) inference."""

    drift_compensation: BaseDriftCompensation = field(default_factory=GlobalDriftCompensation)
    """For compensating the drift during inference only."""

    clip: WeightClipParameter = field(default_factory=WebComposerWeightClipParameter)
    """Parameter for weight clip."""

    modifier: WeightModifierParameter = field(default_factory=WebComposerWeightModifierParameter)
    """Parameter for weight modifier."""

    mapping: MappingParameter = field(default_factory=WebComposerMappingParameter)
    """Parameter related to mapping weights to tiles for supporting modules."""


@dataclass
class OldWebComposerMappingParameter(WebComposerMappingParameter):
    """Parameter related to hardware design and the mapping of logical
    weight matrices to physical tiles.

    The default values used for the web composer.
    """

    digital_bias: bool = False

    weight_scaling_omega: float = 1.0
    weight_scaling_columnwise: bool = True
    learn_out_scaling: bool = False
    out_scaling_columnwise: bool = False


@dataclass
class OldWebComposerInferenceRPUConfig(WebComposerInferenceRPUConfig):
    """Preset configuration used as default for the Inference Composer
    """

    mapping: MappingParameter = field(default_factory=OldWebComposerMappingParameter)
    """Parameter related to mapping weights to tiles for supporting modules."""



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Tests for linear layer."""

from numpy.testing import assert_array_almost_equal

from torch import Tensor, manual_seed
from torch.nn import Sequential, Linear as torchLinear
from torch.nn.functional import mse_loss
from torch.optim import SGD

from aihwkit.optim import AnalogSGD
from aihwkit.simulator.configs.configs import InferenceRPUConfig, FloatingPointRPUConfig
from aihwkit.simulator.configs.utils import (
    MappingParameter, WeightModifierType, WeightModifierParameter, IOParameters,
    WeightRemapType
)
from aihwkit.inference.compensation.drift import GlobalDriftCompensation
from aihwkit.inference.noise.custom import StateIndependentNoiseModel


from aihwkit.nn import AnalogSequential, AnalogLinear

from .helpers.decorators import parametrize_over_layers
from .helpers.layers import Linear, LinearCuda
from .helpers.testcases import ParametrizedTestCase
from .helpers.tiles import FloatingPoint, IdealizedConstantStep, Inference


@parametrize_over_layers(
    layers=[Linear, LinearCuda],
    tiles=[FloatingPoint, IdealizedConstantStep, Inference],
    biases=['analog', 'digital', None]
)
class LinearLayerTest(ParametrizedTestCase):
    """Linear layer abstractions tests."""

    @staticmethod
    def train_model(model, loss_func, x_b, y_b, **kwargs):
        """Train the model."""
        opt = AnalogSGD(model.parameters(), lr=0.5,  **kwargs)
        opt.regroup_param_groups(model)

        epochs = 100
        for _ in range(epochs):
            opt.zero_grad()
            pred = model(x_b)
            loss = loss_func(pred, y_b)

            loss.backward()
            opt.step()

    @staticmethod
    def train_model_torch(model, loss_func, x_b, y_b, **kwargs):
        """Train the model with torch SGD."""
        opt = SGD(model.parameters(), lr=0.5, **kwargs)
        epochs = 100
        for _ in range(epochs):
            opt.zero_grad()
            pred = model(x_b)
            loss = loss_func(pred, y_b)

            loss.backward()
            opt.step()

    def test_single_analog_layer(self):
        """Check using a single layer."""
        loss_func = mse_loss

        x_b = Tensor([[0.1, 0.2], [0.2, 0.4]])
        y_b = Tensor([[0.3], [0.6]])

        manual_seed(4321)
        model = self.get_layer(2, 1)
        if self.use_cuda:
            x_b = x_b.cuda()
            y_b = y_b.cuda()
            model = model.cuda()

        initial_loss = loss_func(model(x_b), y_b)
        self.train_model(model, loss_func, x_b, y_b)
        self.assertLess(loss_func(model(x_b), y_b), initial_loss)

    def test_single_analog_layer_sequential(self):
        """Check using a single layer as a Sequential."""
        loss_func = mse_loss

        x_b = Tensor([[0.1, 0.2], [0.2, 0.4]])
        y_b = Tensor([[0.3], [0.6]])

        manual_seed(4321)
        model = Sequential(self.get_layer(2, 1))
        if self.use_cuda:
            x_b = x_b.cuda()
            y_b = y_b.cuda()
            model = model.cuda()

        initial_loss = loss_func(model(x_b), y_b)
        self.train_model(model, loss_func, x_b, y_b)
        self.assertLess(loss_func(model(x_b), y_b), initial_loss)

    def test_seed(self):
        """Check layer seed."""

        manual_seed(4321)
        layer1 = self.get_layer(4, 2)

        manual_seed(4321)
        layer2 = self.get_layer(4, 2)

        weight1, bias1 = layer1.get_weights()
        weight2, bias2 = layer2.get_weights()

        assert_array_almost_equal(weight1, weight2)
        if bias1 is not None:
            assert_array_almost_equal(bias1, bias2)

    def test_several_analog_layers(self):
        """Check using a several analog layers."""
        loss_func = mse_loss

        x_b = Tensor([[0.1, 0.2, 0.3, 0.4], [0.2, 0.4, 0.3, 0.1]])
        y_b = Tensor([[0.3], [0.6]])

        manual_seed(4321)
        model = Sequential(
            self.get_layer(4, 2),
            self.get_layer(2, 1)
        )
        if self.use_cuda:
            x_b = x_b.cuda()
            y_b = y_b.cuda()
            model = model.cuda()

        initial_loss = loss_func(model(x_b), y_b)
        self.train_model(model, loss_func, x_b, y_b)
        self.assertLess(loss_func(model(x_b), y_b), initial_loss)

    def test_analog_and_digital(self):
        """Check mixing analog and digital layers."""
        loss_func = mse_loss

        x_b = Tensor([[0.1, 0.2, 0.3, 0.4], [0.2, 0.4, 0.3, 0.1]])
        y_b = Tensor([[0.3], [0.6]])

        manual_seed(4321)
        model = Sequential(
            self.get_layer(4, 3),
            torchLinear(3, 3),
            self.get_layer(3, 1)
        )
        if self.use_cuda:
            x_b = x_b.cuda()
            y_b = y_b.cuda()
            model = model.cuda()

        initial_loss = loss_func(model(x_b), y_b)
        self.train_model(model, loss_func, x_b, y_b)
        self.assertLess(loss_func(model(x_b), y_b), initial_loss)

    def test_analog_torch_optimizer(self):
        """Check analog layers with torch SGD for inference."""
        loss_func = mse_loss

        x_b = Tensor([[0.1, 0.2, 0.3, 0.4], [0.2, 0.4, 0.3, 0.1]])
        y_b = Tensor([[0.3], [0.6]])

        manual_seed(4321)
        model = Sequential(
            self.get_layer(4, 3),
            self.get_layer(3, 1),
        )
        if not isinstance(model[0].analog_tile.rpu_config, InferenceRPUConfig):
            return

        manual_seed(4321)
        rpu_config = FloatingPointRPUConfig(
            mapping=MappingParameter(digital_bias=self.digital_bias)
        )
        model2 = AnalogSequential(
            AnalogLinear(4, 3, rpu_config=rpu_config, bias=self.bias),
            AnalogLinear(3, 1, rpu_config=rpu_config, bias=self.bias)
        )

        if self.use_cuda:
            x_b = x_b.cuda()
            y_b = y_b.cuda()
            model = model.cuda()
            model2 = model2.cuda()

        initial_loss = loss_func(model(x_b), y_b)

        # train with SGD
        self.train_model_torch(model, loss_func, x_b, y_b)
        self.assertLess(loss_func(model(x_b), y_b), initial_loss)

        # train with AnalogSGD
        self.train_model(model2, loss_func, x_b, y_b)
        self.assertLess(loss_func(model2(x_b), y_b), initial_loss)
        final_loss = loss_func(model(x_b), y_b).detach().cpu().numpy()
        final_loss2 = loss_func(model2(x_b), y_b).detach().cpu().numpy()

        assert_array_almost_equal(final_loss, final_loss2)

    def test_learning_rate_update(self):
        """Check the learning rate update is applied to tile."""
        loss_func = mse_loss

        x_b = Tensor([[0.1, 0.2], [0.2, 0.4]])
        y_b = Tensor([[0.3], [0.6]])

        layer1 = self.get_layer(2, 3)
        layer2 = self.get_layer(3, 1)

        model = Sequential(layer1, layer2)
        if self.use_cuda:
            x_b = x_b.cuda()
            y_b = y_b.cuda()
            model = model.cuda()
        opt = AnalogSGD(model.parameters(), lr=0.5)
        opt.regroup_param_groups(model)
        opt.zero_grad()

        new_lr = 0.07
        for param_group in opt.param_groups:
            param_group['lr'] = new_lr

        pred = model(x_b)
        loss = loss_func(pred, y_b)
        loss.backward()
        opt.step()

        if not layer1.analog_tile.get_analog_ctx().use_torch_update:
            self.assertAlmostEqual(layer1.analog_tile.get_learning_rate(), new_lr)

    def test_learning_rate_update_fn(self):
        """Check the learning rate update is applied to tile."""
        layer1 = self.get_layer(2, 3)
        layer2 = self.get_layer(3, 1)

        model = Sequential(layer1, layer2)
        if self.use_cuda:
            model = model.cuda()
        opt = AnalogSGD(model.parameters(), lr=0.5)
        opt.regroup_param_groups(model)
        opt.zero_grad()

        new_lr = 0.07

        opt.set_learning_rate(new_lr)

        self.assertAlmostEqual(layer1.analog_tile.get_learning_rate(), new_lr)
        self.assertAlmostEqual(layer2.analog_tile.get_learning_rate(), new_lr)

    def test_out_scaling_learning(self):
        """Check if out scales are learning."""
        loss_func = mse_loss

        x_b = Tensor([[0.1, 0.2, 0.3, 0.4], [0.2, 0.4, 0.3, 0.1]])
        y_b = Tensor([[0.3], [0.6]])

        manual_seed(4321)

        rpu_config = InferenceRPUConfig(mapping=MappingParameter(
            weight_scaling_omega=0.6,
            learn_out_scaling=True,
            out_scaling_columnwise=False))

        model = Sequential(
            self.get_layer(4, 2, rpu_config=rpu_config),
            self.get_layer(2, 1, rpu_config=rpu_config)
        )
        if self.use_cuda:
            x_b = x_b.cuda()
            y_b = y_b.cuda()
            model = model.cuda()

        initial_out_scaling_0 = model[0].analog_tile.get_learned_out_scales().clone()
        initial_out_scaling_1 = model[1].analog_tile.get_learned_out_scales().clone()

        self.train_model(model, loss_func, x_b, y_b)

        learned_out_scaling_0 = model[0].analog_tile.get_learned_out_scales().data.clone()
        learned_out_scaling_1 = model[1].analog_tile.get_learned_out_scales().data.clone()

        self.assertEqual(initial_out_scaling_0.numel(), 1)
        self.assertIsNotNone(model[0].analog_tile.get_learned_out_scales().grad)
        self.assertNotAlmostEqualTensor(initial_out_scaling_0, learned_out_scaling_0)

        self.assertEqual(initial_out_scaling_0.numel(), 1)
        self.assertIsNotNone(model[1].analog_tile.get_learned_out_scales().grad)
        self.assertNotAlmostEqualTensor(initial_out_scaling_1, learned_out_scaling_1)
        self.assertEqual(initial_out_scaling_0.numel(), 1)

    def test_out_scaling_learning_columnwise(self):
        """Check if out scaling alpha are learning when columnwise is True."""
        loss_func = mse_loss

        x_b = Tensor([[0.1, 0.2, 0.3, 0.4], [0.2, 0.4, 0.3, 0.1]])
        y_b = Tensor([[0.3], [0.6]])

        manual_seed(4321)

        rpu_config = InferenceRPUConfig(mapping=MappingParameter(
            weight_scaling_omega=0.6,
            learn_out_scaling=True,
            weight_scaling_columnwise=True))

        model = Sequential(
            self.get_layer(4, 2, rpu_config=rpu_config),
            self.get_layer(2, 1, rpu_config=rpu_config)
        )
        if self.use_cuda:
            x_b = x_b.cuda()
            y_b = y_b.cuda()
            model = model.cuda()

        initial_out_scaling_0 = model[0].analog_tile.get_learned_out_scales().clone()
        initial_out_scaling_1 = model[1].analog_tile.get_learned_out_scales().clone()

        self.train_model(model, loss_func, x_b, y_b)

        learned_out_scaling_0 = model[0].analog_tile.get_learned_out_scales().clone()
        learned_out_scaling_1 = model[1].analog_tile.get_learned_out_scales().clone()

        self.assertGreaterEqual(initial_out_scaling_0.numel(), 1)
        self.assertIsNotNone(model[0].analog_tile.get_learned_out_scales().grad)
        self.assertNotAlmostEqualTensor(initial_out_scaling_0, learned_out_scaling_0)

        self.assertGreaterEqual(initial_out_scaling_1.numel(), 1)
        self.assertIsNotNone(model[1].analog_tile.get_learned_out_scales().grad)
        self.assertNotAlmostEqualTensor(initial_out_scaling_1, learned_out_scaling_1)


@parametrize_over_layers(
    layers=[Linear, LinearCuda],
    tiles=[Inference],
    biases=['digital']
)
class LinearLayerInferenceTest(ParametrizedTestCase):
    """Linear layer abstractions tests for inference."""

    def test_remapping_learning(self):
        """Check analog layers with torch SGD for inference."""
        loss_func = mse_loss

        x_b = Tensor([[0.1, 0.2, 0.3, 0.4], [0.2, 0.4, 0.3, 0.1]])
        y_b = Tensor([[0.3], [0.6]])

        manual_seed(4321)
        rpu_config = self.get_rpu_config()
        rpu_config.remap.type = WeightRemapType.LAYERWISE_SYMMETRIC
        rpu_config.forward.is_perfect = True

        rpu_config.mapping.learn_out_scaling = False
        rpu_config.mapping.weight_scaling_omega = 1.0
        rpu_config.mapping.weight_scaling_columnwise = False

        analog_model = Sequential(
            self.get_layer(4, 3, rpu_config=rpu_config),
            self.get_layer(3, 1, rpu_config=rpu_config),
        )

        manual_seed(4321)
        torch_model = Sequential(
            torchLinear(4, 3, bias=self.bias),
            torchLinear(3, 1, bias=self.bias)
        )

        if self.use_cuda:
            x_b = x_b.cuda()
            y_b = y_b.cuda()
            analog_model = analog_model.cuda()
            torch_model = torch_model.cuda()

        initial_loss = loss_func(analog_model(x_b), y_b).detach().cpu().numpy()

        # train analog model with AnalogSGD
        LinearLayerTest.train_model(analog_model, loss_func, x_b, y_b)
        self.assertLess(loss_func(analog_model(x_b), y_b).detach().cpu().numpy(),
                        initial_loss)

        # check remapping
        tile_weights = analog_model[0].get_weights(apply_weight_scaling=False)[0]
        self.assertAlmostEqual(tile_weights.abs().flatten().max().item(), 1.0)

        # train torch model with SGD
        initial_loss2 = loss_func(torch_model(x_b), y_b).detach().cpu().numpy()
        LinearLayerTest.train_model_torch(torch_model, loss_func, x_b, y_b)
        self.assertLess(loss_func(torch_model(x_b), y_b).detach().cpu().numpy(),
                        initial_loss)

        # should be same
        final_loss = loss_func(analog_model(x_b), y_b).detach().cpu().numpy()
        final_loss2 = loss_func(torch_model(x_b), y_b).detach().cpu().numpy()

        assert_array_almost_equal(initial_loss, initial_loss2)
        assert_array_almost_equal(final_loss, final_loss2)

    def test_inference_modifier(self):
        """ tests whether modifier are used """

        x_b = Tensor([[0.1, 0.2, 0.3, 0.4], [0.2, 0.4, 0.3, 0.1]])

        rpu_config = InferenceRPUConfig(
            mapping=MappingParameter(
                weight_scaling_omega=0.0,
                learn_out_scaling=False,
                weight_scaling_columnwise=False
            ),
            modifier=WeightModifierParameter(
                type=WeightModifierType.ADD_NORMAL,
                std_dev=1.0,
            ),
            forward=IOParameters(is_perfect=True)
        )

        model = AnalogSequential(
            self.get_layer(4, 2, rpu_config=rpu_config),
            self.get_layer(2, 1, rpu_config=rpu_config)
        )
        if self.use_cuda:
            x_b = x_b.cuda()
            model = model.cuda()

        opt = AnalogSGD(model.parameters(), lr=0.0)
        opt.step()

        model.eval()
        y_eval1 = model(x_b)
        model.train()
        opt.step()
        model.eval()
        y_eval2 = model(x_b)

        model.train()
        y_train1 = model(x_b)
        opt.step()
        y_train2 = model(x_b)

        self.assertNotAlmostEqualTensor(y_train1, y_train2)
        self.assertTensorAlmostEqual(y_eval2, y_eval1)

    def test_drift_compensation(self):
        """ tests whether drift compensation is performed """

        x_b = Tensor([[0.1, 0.2, 0.3, 0.4], [0.2, 0.4, 0.3, 0.1]])

        rpu_config = InferenceRPUConfig(
            mapping=MappingParameter(
                weight_scaling_omega=0.0,
                learn_out_scaling=False,
                weight_scaling_columnwise=False
            ),
            forward=IOParameters(is_perfect=True),
            drift_compensation=GlobalDriftCompensation(),
            noise_model=StateIndependentNoiseModel(
                prog_noise_scale=0.0,
                read_noise_scale=0.0,
                drift_nu_std=0.0,
                drift_nu_mean=0.1,
            )
        )

        model = AnalogSequential(
            self.get_layer(4, 2, rpu_config=rpu_config),
            self.get_layer(2, 1, rpu_config=rpu_config)
        )

        rpu_config.drift_compensation = None
        model_without = AnalogSequential(
            self.get_layer(4, 2, rpu_config=rpu_config),
            self.get_layer(2, 1, rpu_config=rpu_config)
        )

        model_without.load_state_dict(model.state_dict(), load_rpu_config=False)

        if self.use_cuda:
            x_b = x_b.cuda()
            model = model.cuda()
            model_without = model_without.cuda()

        model.eval()
        y_before = model(x_b)
        model.drift_analog_weights(1000.)
        y_after = model(x_b)

        model_without.eval()
        y_without_before = model_without(x_b)
        model_without.drift_analog_weights(1000.)
        y_without_after = model_without(x_b)

        self.assertTensorAlmostEqual(y_before, y_without_before)
        self.assertTensorAlmostEqual(y_before, y_after)
        self.assertNotAlmostEqualTensor(y_after, y_without_after)
        self.assertNotAlmostEqualTensor(y_without_before, y_without_after)



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""High level analog tiles (inference)."""

from typing import List, Optional, Union, Any, TYPE_CHECKING

from torch import device as torch_device
from torch import ones, zeros, Tensor, float32
from torch.autograd import no_grad

from aihwkit.simulator.tiles.analog import AnalogTile

if TYPE_CHECKING:
    from aihwkit.simulator.configs import InferenceRPUConfig
    from aihwkit.simulator.tiles import BaseTile

# pylint: disable=too-many-instance-attributes


class InferenceTile(AnalogTile):
    """Tile used for analog inference and hardware-aware training for inference.

    Args:
        out_size: output size
        in_size: input size
        rpu_config: resistive processing unit configuration.
        bias: whether to add a bias column to the tile.
        in_trans: Whether to assume an transposed input (batch first)
        out_trans: Whether to assume an transposed output (batch first)
        shared_weights: Whether to keep the weight in torch's memory space
    """

    def __init__(
            self,
            out_size: int,
            in_size: int,
            rpu_config: Optional['InferenceRPUConfig'] = None,
            bias: bool = False,
            in_trans: bool = False,
            out_trans: bool = False,
            shared_weights: bool = True,
    ):

        if not rpu_config:
            # Import `InferenceRPUConfig` dynamically to avoid import cycles.
            # pylint: disable=import-outside-toplevel
            from aihwkit.simulator.configs import InferenceRPUConfig
            rpu_config = InferenceRPUConfig()

        # CAUTION: one cannot save parts of the RPUConfig as
        # properties of the class! This is because those attributes
        # would be restored even if the RPUConfig would be replaced
        # during checkpoint loading

        self.drift_baseline = None
        self.drift_readout_tensor = None  # type: Optional[Tensor]
        self.alpha = ones((1,))

        # Helpers.
        self.programmed_weights = None  # type: Optional[Tensor]
        self.nu_drift_list = None  # type: Optional[List[Tensor]]

        super().__init__(out_size, in_size, rpu_config, bias, in_trans, out_trans)

        if shared_weights:
            self.shared_weights = zeros(out_size, in_size + int(bias),
                                        requires_grad=True)  # type: Tensor
            self.ensure_shared_weights()

    @no_grad()
    def init_mapping_scales(self) -> None:
        """Helper function to initialize the mapping scales used to scale the
        weights in digital and determine the conductance conversion.

        Note:
            This method is called from the constructor.
        """
        # Import `aihwkit.simulator.configs` items dynamically to avoid import cycles.
        # pylint: disable=import-outside-toplevel
        from aihwkit.simulator.configs.utils import WeightRemapType

        super().init_mapping_scales()
        remap = self.rpu_config.remap  # type: ignore
        if remap.type != WeightRemapType.NONE:
            # needs to be always out_size
            mapping_scales = ones((self.out_size, ),
                                  dtype=float32,
                                  device=self.device,
                                  requires_grad=False)
            self.set_mapping_scales(mapping_scales)

    @no_grad()
    def _forward_drift_readout_tensor(self, reset_if: bool = False) -> Optional[Tensor]:
        """Perform a forward pass using the drift read-out tensor.

        Args:
            reset_if: Will reset the readout tensor, otherwise use the stored one

        Returns:
            Readout tensor if drift compensation is on
        """

        if self.rpu_config.drift_compensation is None:
            return None

        if self.drift_readout_tensor is None or reset_if:
            self.drift_readout_tensor = self.rpu_config.drift_compensation.get_readout_tensor(
                self.tile.get_x_size()).detach().to(self.device)
            if self.in_trans:
                self.drift_readout_tensor = self.drift_readout_tensor.tranpose(0, 1).clone()

        # We need to take the bias as a common column here, also we do
        # not want to use indexed.
        return self.tile.forward(self.drift_readout_tensor, False,
                                 self.in_trans, self.out_trans, True)

    @no_grad()
    def program_weights(self, from_reference: bool = True) -> None:
        """Apply weights noise to the current tile weights and saves these for
        repeated drift experiments.

        This method also establishes the drift coefficients for each
        conductance slice.

        Will also reset the drift readout tensor and compuate a new
        drift compensation baseline

        Args:
            from_reference: Whether to use weights from reference
        """
        # pylint: disable=arguments-differ

        if not from_reference or self.reference_combined_weights is None:
            self.reference_combined_weights = Tensor(self.tile.get_weights())

        self.programmed_weights, self.nu_drift_list = \
            self.rpu_config.noise_model.apply_programming_noise(
                self.reference_combined_weights)

        self.tile.set_weights(self.programmed_weights)

        if self.rpu_config.drift_compensation is not None:
            forward_output = self._forward_drift_readout_tensor(True)
            self.drift_baseline = self.rpu_config.drift_compensation.init_baseline(forward_output)

    @no_grad()
    def drift_weights(
            self,
            t_inference: float = 0.0
    ) -> None:
        """Programs and drifts the current reference weights.

        The current weight reference is either the current weights or
        the ones at the time when :meth:`initialize_drift_reference`
        was called, which then would overwrite the current weights
        with the drifted ones.

        Args:
            t_inference: Time (in sec) of assumed inference
                time. Programming ends at t=0s.  The rest is waiting time,
                where the devices might drift and accumulate noise. See
                noise model used for details.
        """
        # pylint: disable=arguments-differ, arguments-renamed

        if self.programmed_weights is None:
            self.program_weights()

        drifted_weights = self.rpu_config.noise_model.apply_drift_noise(
            self.programmed_weights, self.nu_drift_list, t_inference)
        self.tile.set_weights(drifted_weights)

        if self.rpu_config.drift_compensation is not None:
            forward_output = self._forward_drift_readout_tensor()
            self.alpha = self.rpu_config.drift_compensation.apply(
                forward_output,
                self.drift_baseline).to(self.device)

    @no_grad()
    def post_forward(self, x_output: Tensor, dim: int, is_test: bool = False,
                     ctx: Any = None) -> Tensor:
        """Operations after the actual forward step for post processing """

        x_output = super().post_forward(x_output, dim, is_test, ctx)

        if not is_test or self.rpu_config.drift_compensation is None:
            return x_output

        # only do drift compensation in eval mode
        return x_output * self.alpha

    @no_grad()
    def post_update_step(self) -> None:
        """Operators that need to be called once per mini-batch.

        In the :class:`~InferenceTile`, the following calls are made
        (if enabled in the ``rpu_config`` settings). First, the post
        update step of the parent is called, then the weight clipping
        is done, subsequently then remapping is done (if enforced),
        and finally the forward-backward weight modifier is
        called. The latter will modify the weights that are used
        during forward and backward (but not update) until the next
        time this function is called.

        """
        # Import `aihwkit.simulator.configs` items dynamically to avoid import cycles.
        # pylint: disable=import-outside-toplevel
        from aihwkit.simulator.configs.helpers import parameters_to_bindings
        from aihwkit.simulator.configs.utils import (
            WeightClipType, WeightModifierType, WeightRemapType
        )

        super().post_update_step()

        # TODO: make this a little nicer. Now each time bindings are
        # generated, which however has the advantage that parameters
        # could be changed-on-the-fly

        if self.rpu_config.clip.type != WeightClipType.NONE:
            weight_clip_params = parameters_to_bindings(self.rpu_config.clip)
            self.tile.clip_weights(weight_clip_params)

        if self.rpu_config.remap.type != WeightRemapType.NONE:
            weight_remap_params = parameters_to_bindings(self.rpu_config.remap)
            scales = self.get_scales()
            scales = self.tile.remap_weights(weight_remap_params, scales)
            self.set_scales(scales)

        # update the forward / backward modified weights here
        if (self.rpu_config.modifier.type != WeightModifierType.COPY or
                self.rpu_config.modifier.pdrop > 0.0):
            weight_modify_params = parameters_to_bindings(self.rpu_config.modifier)
            self.tile.modify_weights(weight_modify_params)

    def cuda(
            self,
            device: Optional[Union[torch_device, str, int]] = None
    ) -> 'BaseTile':
        """Return a copy of this tile in CUDA memory.

        Args:
            device: CUDA device

        Returns:
            Self with the underlying C++ tile moved to CUDA memory.

        Raises:
            CudaError: if the library has not been compiled with CUDA.
        """
        super().cuda(device)

        self.alpha = self.alpha.cuda(device)
        self.shared_weights.data = zeros(self.tile.get_x_size(),
                                         self.tile.get_d_size(),
                                         requires_grad=True).cuda(device)
        self.ensure_shared_weights()

        return self



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Base class for analog Modules."""
from typing import (
    Any, Dict, List, Optional, Tuple, NamedTuple, Union,
    Generator, TYPE_CHECKING
)
from copy import deepcopy
from warnings import warn

from torch import Tensor, no_grad
from torch.nn import Module, Parameter
from torch import device as torch_device

from aihwkit.exceptions import ModuleError
from aihwkit.simulator.configs.utils import MappingParameter
from aihwkit.simulator.tiles import InferenceTile
from aihwkit.optim.context import AnalogContext

if TYPE_CHECKING:
    from aihwkit.simulator.tiles import BaseTile
    from collections import OrderedDict
    from aihwkit.simulator.configs.configs import (
        FloatingPointRPUConfig, InferenceRPUConfig, SingleRPUConfig,
        UnitCellRPUConfig, DigitalRankUpdateRPUConfig
    )

RPUConfigAlias = Union['FloatingPointRPUConfig', 'SingleRPUConfig',
                       'UnitCellRPUConfig', 'InferenceRPUConfig',
                       'DigitalRankUpdateRPUConfig']


class AnalogModuleBase(Module):
    """Base class for analog Modules.

    Base ``Module`` for analog layers that use analog tiles. When subclassing,
    please note:

    * the :meth:`_setup_tile()` method is expected to be called by the subclass
      constructor, and it does not only create a tile.
    * :meth:`register_analog_tile` needs to be called for each created analog tile
    * this module does *not* call torch's ``Module`` init as the child is
      likely again derived from Module
    * the ``weight`` and ``bias`` Parameters are not guaranteed to be in
      sync with the tile weights and biases during the lifetime of the instance,
      for performance reasons. The canonical way of reading and writing
      weights is via the :meth:`set_weights()` and :meth:`get_weights()` as opposed
      to using the attributes directly.
    * the ``BaseTile`` subclass that is created is retrieved from the
      ``rpu_config.tile_class`` attribute.

    Args:
        in_features: input vector size (number of columns).
        out_features: output vector size (number of rows).
        bias: whether to use a bias row on the analog tile or not.
        realistic_read_write: whether to enable realistic read/write
            for setting initial weights and during reading of the weights.
        mapping: Configuration of the hardware architecture (e.g. tile size).
    """
    # pylint: disable=abstract-method, too-many-instance-attributes
    ANALOG_CTX_PREFIX: str = 'analog_ctx_'
    ANALOG_SHARED_WEIGHT_PREFIX: str = 'analog_shared_weights_'
    ANALOG_STATE_PREFIX: str = 'analog_tile_state_'
    ANALOG_OUT_SCALING_ALPHA_PREFIX: str = 'analog_out_scaling_alpha_'
    ANALOG_INPUT_RANGE_PREFIX: str = 'analog_input_range_'

    def __init__(
            self,
            in_features: int,
            out_features: int,
            bias: bool,
            realistic_read_write: bool = False,
            mapping: Optional[MappingParameter] = None,
    ) -> None:
        # pylint: disable=super-init-not-called
        self._analog_tile_counter = 0
        self._registered_helper_parameter = []  # type: list
        self._load_rpu_config = True
        self._strict_rpu_config_check = True

        if mapping is None:
            mapping = MappingParameter()

        self.use_bias = bias
        self.digital_bias = bias and mapping.digital_bias
        self.analog_bias = bias and not mapping.digital_bias
        self.realistic_read_write = realistic_read_write
        self.in_features = in_features
        self.out_features = out_features

    def register_helper(self, name: str) -> None:
        """Register a helper name that is not saved to the state dict """

        if name not in self._registered_helper_parameter:
            self._registered_helper_parameter.append(name)

    def register_analog_tile(self, tile: 'BaseTile', name: Optional[str] = None,
                             update_only: bool = False) -> None:
        """Register the analog context of the tile.

        Note:
            Needs to be called at the end init to register the tile
            for the analog optimizers.

        Args:
            tile: tile to register
            name: Optional tile name used as the parameter name.
            update_only: Whether to re-register (does not advance tile counter)
        """

        if name is None:
            name = str(self._analog_tile_counter)

        if not update_only:
            self._analog_tile_counter += 1

        ctx_name = self.ANALOG_CTX_PREFIX + name

        self.register_helper(ctx_name)
        self.register_parameter(ctx_name, tile.get_analog_ctx())

        if tile.shared_weights is not None:
            if not isinstance(tile.shared_weights, Parameter):
                tile.shared_weights = Parameter(tile.shared_weights)
            par_name = self.ANALOG_SHARED_WEIGHT_PREFIX + name
            self.register_parameter(par_name, tile.shared_weights)
            self.register_helper(par_name)

        if tile.out_scaling_alpha is not None:
            if not isinstance(tile.out_scaling_alpha, Parameter):
                tile.out_scaling_alpha = Parameter(tile.out_scaling_alpha)
            par_name = self.ANALOG_OUT_SCALING_ALPHA_PREFIX + name
            self.register_parameter(par_name, tile.out_scaling_alpha)
            self.register_helper(par_name)

        if tile.input_range is not None:
            if not isinstance(tile.input_range, Parameter):
                tile.input_range = Parameter(tile.input_range)
            par_name = self.ANALOG_INPUT_RANGE_PREFIX + name
            self.register_parameter(par_name, tile.input_range)
            self.register_helper(par_name)

    def unregister_parameter(self, param_name: str) -> None:
        """Unregister module parameter from parameters.

        Raises:
            ModuleError: In case parameter is not found
        """
        param = getattr(self, param_name, None)
        if not isinstance(param, Parameter):
            raise ModuleError(f"Cannot find parameter {param_name} to unregister")
        param_data = param.detach().clone()
        delattr(self, param_name)
        setattr(self, param_name, param_data)

    def analog_tiles(self) -> Generator['BaseTile', None, None]:
        """ Generator to loop over all registered analog tiles of the module """
        for param in self.parameters():
            if isinstance(param, AnalogContext):
                yield param.analog_tile

    def named_analog_tiles(self) -> Generator[Tuple[str, 'BaseTile'], None, None]:
        """ Generator to loop over all registered analog tiles of the module with names. """
        for name, param in self.named_parameters():
            if isinstance(param, AnalogContext):
                new_name = name.split(self.ANALOG_CTX_PREFIX)[-1]
                yield (new_name, param.analog_tile)

    def get_analog_tile_devices(self) -> List[Optional[Union[torch_device, str, int]]]:
        """ Return a list of the devices used by the analog tiles.

        Returns:
            List of torch devices
        """
        return [d.device for d in self.analog_tiles()]

    def analog_tile_count(self) -> int:
        """Return the number of registered tiles.

        Returns:
           Number of registered tiles

        """
        return getattr(self, '_analog_tile_counter', 0)

    def _setup_tile(
            self,
            rpu_config: RPUConfigAlias,
    ) -> 'BaseTile':
        """Create a single analog tile with the given RPU configuration.

        Create an analog tile to be used for the basis of this layer
        operations, while using the attributes ``(in_features,
        out_features, bias)`` given to this instance during init.

        After tile creation, the tile needs to be registered using
        :meth:`register_analog_tile`.

        Args:
            rpu_config: resistive processing unit configuration.

        Returns:
            An analog tile with the requested parameters.

        """
        # pylint: disable=protected-access

        # Create the tile.
        return rpu_config.tile_class(self.out_features, self.in_features, rpu_config,
                                     bias=self.analog_bias)

    def set_weights(
            self,
            weight: Tensor,
            bias: Optional[Tensor] = None,
            force_exact: bool = False,
            apply_weight_scaling: bool = True,
            weight_scaling_omega: Optional[float] = None
    ) -> None:
        """Set the weight (and bias) values with given tensors.

        This uses an realistic write if the property ``realistic_read_write``
        of the layer is set, unless it is overwritten by ``force_exact``.

        If ``weight_scaling_omega`` is larger than 0, the weights are set in a
        scaled manner (assuming a digital output scale). See
        :meth:`~aihwkit.simulator.tiles.base.apply_weight_scaling`
        for details.

        Note:
            This is the recommended way for setting the weight/bias matrix of
            the analog tile, as it will correctly store the weights into the
            internal memory. Directly writing to ``self.weight`` and
            ``self.bias`` might yield wrong results as they are not always in
            sync with the analog tile Parameters, for performance reasons.

        Args:
            weight: weight matrix
            bias: bias vector
            force_exact: forces an exact write to the analog tiles
            apply_weight_scaling: Whether to rescale the given weight matrix
                and populate the digital output scaling factors as
                specified in the configuration
                :class:`~aihwkit.configs.utils.MappingParameter`. A
                new ``weight_scaling_omega`` can be given. Note that
                this will overwrite the existing digital out scaling
                factors.
            weight_scaling_omega: The weight scaling omega factor (see
                :class:`~aihwkit.configs.utils.MappingParameter`). If
                given explicitly here, it will overwrite the value in
                the mapping field.

        Raises:
            ModuleError: in case of multiple defined analog tiles in the module

        """
        analog_tiles = list(self.analog_tiles())
        if len(analog_tiles) != 1:
            raise ModuleError("AnalogModuleBase.set_weights only supports a single tile.")

        shape = [self.out_features, self.in_features]
        weight = weight.clone().reshape(shape)

        analog_tiles[0].set_weights(
            weight, bias if self.analog_bias else None,
            apply_weight_scaling,
            weight_scaling_omega)

        realistic = self.realistic_read_write and not force_exact
        if realistic:
            analog_tiles[0].program_weights()

        if bias is not None and self.digital_bias:
            with no_grad():
                self.bias.data[:] = bias[:]

        self._sync_weights_from_tile()

    def get_weights(
            self,
            force_exact: bool = False,
            apply_weight_scaling: bool = True,
    ) -> Tuple[Tensor, Optional[Tensor]]:
        """Get the weight (and bias) tensors.

        This uses an realistic read if the property ``realistic_read_write`` of
        the layer is set, unless it is overwritten by ``force_exact``. It
        scales the analog weights by the digital output scales by default.

        Note:
            This is the recommended way for setting the weight/bias matrix from
            the analog tile, as it will correctly fetch the weights from the
            internal memory. Accessing ``self.weight`` and ``self.bias`` might
            yield wrong results as they are not always in sync with the
            analog tile library, for performance reasons.

        Args:
            force_exact: Forces an exact read to the analog tiles
            apply_weight_scaling: Whether to return the weights with the
                (digital) output scaling factors applied. Note the
                "logical" weights of the layer which the DNN is
                effectively using are those with the output scales
                applied. If ``apply_weight_scaling`` is set to False, then
                only the weight values that is programmed onto the
                crossbar array are returned, without applying the
                digital scales. Default is True.

        Returns:
            tuple: weight matrix, bias vector

        Raises:
            ModuleError: in case of multiple defined analog tiles in the module

        """
        analog_tiles = list(self.analog_tiles())
        if len(analog_tiles) != 1:
            raise ModuleError("AnalogModuleBase.get_weights only supports a single tile.")

        realistic = self.realistic_read_write and not force_exact
        if realistic:
            weight, analog_bias = analog_tiles[0].read_weights(
                apply_weight_scaling)
        else:
            weight, analog_bias = analog_tiles[0].get_weights(
                apply_weight_scaling)

        digital_bias = None
        if self.digital_bias:
            with no_grad():
                digital_bias = self.bias.data.clone().detach().cpu()

        if (digital_bias is not None) and (analog_bias is not None):
            bias = digital_bias + analog_bias
        elif digital_bias is not None:
            bias = digital_bias
        else:
            bias = analog_bias
        return weight, bias

    def remap_weights(self, weight_scaling_omega: Optional[float] = 1.0) -> None:
        """Gets and re-sets the weights in case of using the weight scaling.

        This re-sets the weights with applied mapping scales, so that
        the weight mapping scales are updated.

        In case of hardware-aware training, this would update the
        weight mapping scales so that the absolute max analog weights
        are set to 1 (as specified in the ``weight_scaling``
        configuration of
        :class:`~aihwkit.configs.utils.MappingParameter`).

        Note:
            By default the weight scaling omega factor is set to 1
            here (overriding any setting in the ``rpu_config``). This
            means that the max weight value is set to 1 internally for
            the analog weights.

        Caution:
            This should typically *not* be called for analog
            training unless realistic_read_write is set. In this case,
            it would perform a full re-write of the weights.

        Args:
            weight_scaling_omega: The weight scaling omega factor (see
                :class:`~aihwkit.configs.utils.MappingParameter`). If
                set to None here, it will take the value in the
                mapping parameters. Default is however 1.0.

        """
        weights, biases = self.get_weights(False, True)
        self.set_weights(weights, biases, False, True, weight_scaling_omega)

    def _sync_weights_from_tile(self) -> None:
        """Update the layer weight and bias from the values on the analog tile.

        Update the ``self.weight`` and ``self.bias`` Parameters with an
        exact copy of the internal analog tile weights.
        """
        tile_weight, tile_bias = self.get_weights(force_exact=True)  # type: Tuple[Tensor, Tensor]

        self.weight.data[:] = tile_weight.reshape(self.weight.shape)
        if self.analog_bias:
            with no_grad():
                self.bias.data[:] = tile_bias.reshape(self.bias.shape)

    def _sync_weights_to_tile(self) -> None:
        """Update the tile values from the layer weights and bias.

        Update the internal tile weights with an exact copy of the values of
        the ``self.weight`` and ``self.bias`` Parameters.
        """
        self.set_weights(self.weight, self.bias if self.analog_bias else None,
                         force_exact=True)

    def _set_load_rpu_config_state(self, load_rpu_config: bool = True,
                                   strict_rpu_config_check: bool = True) -> None:
        self._load_rpu_config = load_rpu_config
        self._strict_rpu_config_check = strict_rpu_config_check

    def load_state_dict(self,  # pylint: disable=arguments-differ
                        state_dict: 'OrderedDict[str, Tensor]',
                        strict: bool = True,
                        load_rpu_config: bool = True,
                        strict_rpu_config_check: bool = True) -> NamedTuple:
        """Specializes torch's ``load_state_dict`` to add a flag whether to
        load the RPU config from the saved state.

        Args:
            state_dict: see torch's ``load_state_dict``
            strict: see torch's ``load_state_dict``
            load_rpu_config: Whether to load the saved RPU
                config or use the current RPU config of the model.

                Caution:

                    If ``load_rpu_config=False`` the RPU config can
                    be changed from the stored model. However, the user has to
                    make sure that the changed RPU config makes sense.

                    For instance, changing the device type might
                    change the expected fields in the hidden
                    parameters and result in an error.

            strict_rpu_config_check: Whether to check and throw an
                error if the current ``rpu_config`` is not of the same
                class type when setting ``load_rpu_config`` to
                False. In case of ``False`` the user has to make sure
                that the ``rpu_config`` are compatible.

        Returns:
            see torch's ``load_state_dict``

        Raises:
            ModuleError: in case the rpu_config class mismatches
            or mapping parameter mismatch for
            ``load_rpu_config=False``

        """
        self._set_load_rpu_config_state(load_rpu_config, strict_rpu_config_check)
        return super().load_state_dict(state_dict, strict)

    def __setstate__(self, state: Dict) -> None:
        """Set the state after unpickling.

        Makes sure that the parameter in the tiles are correctly registered.

        """
        self.__dict__.update(state)

        # update registered parameters
        for name, analog_tile in list(self.named_analog_tiles()):
            self.register_analog_tile(analog_tile, name, update_only=True)

    def _load_from_state_dict(
            self,
            state_dict: Dict,
            prefix: str,
            local_metadata: Dict,
            strict: bool,
            missing_keys: List[str],
            unexpected_keys: List[str],
            error_msgs: List[str]) -> None:
        """Copy parameters and buffers from `state_dict` into only this
        module, but not its descendants.

        This method is a specialization of ``Module._load_from_state_dict``
        that takes into account the extra ``analog_tile_state`` key used by
        analog layers.

        Raises:
            ModuleError: in case the rpu_config class mismatches.
        """
        # pylint: disable=too-many-locals, too-many-branches

        for name, analog_tile in list(self.named_analog_tiles()):
            key = prefix + self.ANALOG_STATE_PREFIX + name
            if key not in state_dict:  # legacy
                key = prefix + 'analog_tile_state'

            if key in state_dict:
                analog_state = state_dict.pop(key).copy()

                if not self._load_rpu_config:

                    if self._strict_rpu_config_check:
                        if not isinstance(analog_tile.rpu_config,
                                          type(analog_state['rpu_config'])):
                            raise ModuleError("RPU config mismatch during loading: "
                                              "Tried to replace "
                                              f"{analog_state['rpu_config'].__class__.__name__} "
                                              f"with {analog_tile.rpu_config.__class__.__name__}")

                    if hasattr(analog_state['rpu_config'], 'mapping'):
                        old_mapping = analog_state['rpu_config'].mapping
                        new_mapping = analog_tile.rpu_config.mapping
                        if (old_mapping.max_input_size != new_mapping.max_input_size
                                or old_mapping.max_output_size != new_mapping.max_output_size
                                or old_mapping.digital_bias != new_mapping.digital_bias
                                or (old_mapping.out_scaling_columnwise
                                    != new_mapping.out_scaling_columnwise)):
                            raise ModuleError("MappingParameter mismatch during loading: "
                                              "Tried to replace "
                                              f"{old_mapping} "
                                              f"with {new_mapping}")

                    analog_state['rpu_config'] = analog_tile.rpu_config
                analog_tile.__setstate__(analog_state)

                # update registered parameters
                self.register_analog_tile(analog_tile, name, update_only=True)

            elif strict:
                missing_keys.append(key)

        # update the weight / analog bias (not saved explicitly)
        self._sync_weights_from_tile()

        # remove helper parameters.
        rm_keys = []
        for par_name in self._registered_helper_parameter:
            key = prefix + par_name
            if key in state_dict:
                rm_keys.append(key)

        # legacy
        for part in [self.ANALOG_SHARED_WEIGHT_PREFIX,
                     self.ANALOG_OUT_SCALING_ALPHA_PREFIX]:
            for key in state_dict:
                if part in key:
                    rm_keys.append(key)

        for key in rm_keys:
            if key in state_dict:
                state_dict.pop(key)

        super()._load_from_state_dict(
           state_dict, prefix, local_metadata, strict, missing_keys,
           unexpected_keys, error_msgs)

        # legacy
        for part in [self.ANALOG_SHARED_WEIGHT_PREFIX,
                     self.ANALOG_OUT_SCALING_ALPHA_PREFIX]:
            for key in missing_keys:
                if part in key:
                    rm_keys.append(key)

        # remove the missing keys of the helper parameters
        for key in rm_keys:
            if key in missing_keys:
                missing_keys.remove(key)

    def state_dict(  # pylint: disable=arguments-differ
            self,
            destination: Any = None,
            prefix: str = '',
            keep_vars: bool = False
    ) -> Dict:
        """Return a dictionary containing a whole state of the module."""
        self._sync_weights_from_tile()

        current_state = super().state_dict(destination=destination,
                                           prefix=prefix,
                                           keep_vars=keep_vars)

        for name, analog_tile in self.named_analog_tiles():
            analog_state = analog_tile.__getstate__()
            analog_state_name = prefix + self.ANALOG_STATE_PREFIX + name
            current_state[analog_state_name] = analog_state

        return current_state

    def drift_analog_weights(self, t_inference: float = 0.0) -> None:
        """(Program) and drift the analog weights.

        Args:
            t_inference: assumed time of inference (in sec)

        Raises:
            ModuleError: if the layer is not in evaluation mode.
        """
        if self.training:
            raise ModuleError('drift_analog_weights can only be applied in '
                              'evaluation mode')
        for analog_tile in self.analog_tiles():
            if isinstance(analog_tile, InferenceTile):
                analog_tile.drift_weights(t_inference)

    def program_analog_weights(self) -> None:
        """Program the analog weights.

        Raises:
            ModuleError: if the layer is not in evaluation mode.
        """
        if self.training:
            raise ModuleError('program_analog_weights can only be applied in '
                              'evaluation mode')
        for analog_tile in self.analog_tiles():
            analog_tile.program_weights()

    def extra_repr(self) -> str:
        """Set the extra representation of the module.

        Returns:
            A string with the extra representation.
        """
        output = super().extra_repr()
        if self.realistic_read_write:
            output += ', realistic_read_write={}'.format(self.realistic_read_write)
        if self.analog_bias:
            output += ', analog bias'
        if self.digital_bias:
            output += ', digital bias'

        return output

    def _set_weight_scaling_omega(self, rpu_config: RPUConfigAlias,
                                  weight_scaling_omega: Optional[float] = None
                                  ) -> RPUConfigAlias:
        """ Sets the weight scaling omega and raises a FutureWarning.

        Args:
            rpu_config: dtto
            weight_scaling_omega: parameter value to set

        Returns:
            Copy of ``rpu_config`` if ``weight_scaling_omega`` is not None

        Raises:
            FutureWarning if ``weight_scaling_omega`` is not None
        """

        if weight_scaling_omega is not None:
            warn(FutureWarning("weight_scaling_omega argument to the layer module "
                               "construction will be removed in future. "
                               "Use aihwkit.simulator.configs.utils.MappingParameter "
                               "instead to specify weight scaling."))
            rpu_config = deepcopy(rpu_config)
            rpu_config.mapping.weight_scaling_omega = weight_scaling_omega

        return rpu_config



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Configurations for resistive processing units."""

from dataclasses import dataclass, field
from typing import ClassVar, Type, Optional

from aihwkit.simulator.configs.devices import (
    ConstantStepDevice, FloatingPointDevice, IdealDevice, PulsedDevice
)
from aihwkit.simulator.configs.compounds import (
    DigitalRankUpdateCell, UnitCell,
)
from aihwkit.simulator.configs.helpers import (
    _PrintableMixin, tile_parameters_to_bindings
)
from aihwkit.simulator.configs.utils import (
    IOParameters, PulseType, UpdateParameters, WeightClipParameter,
    WeightModifierParameter, WeightRemapParameter,
    MapableRPU, PrePostProcessingRPU
)
from aihwkit.inference import (
    BaseDriftCompensation, BaseNoiseModel, GlobalDriftCompensation,
    PCMLikeNoiseModel
)
from aihwkit.simulator.rpu_base import devices
from aihwkit.simulator.tiles import AnalogTile, FloatingPointTile, InferenceTile


@dataclass
class FloatingPointRPUConfig(MapableRPU, PrePostProcessingRPU, _PrintableMixin):
    """Configuration for a floating point resistive processing unit."""

    tile_class: ClassVar[Type] = FloatingPointTile
    """Tile class that correspond to this RPUConfig."""

    device: FloatingPointDevice = field(default_factory=FloatingPointDevice)
    """Parameter that modify the behavior of the pulsed device."""


@dataclass
class SingleRPUConfig(MapableRPU, PrePostProcessingRPU, _PrintableMixin):
    """Configuration for an analog (pulsed device) resistive processing unit."""

    tile_class: ClassVar[Type] = AnalogTile
    """Tile class that correspond to this RPUConfig."""

    bindings_class: ClassVar[Type] = devices.AnalogTileParameter

    device: PulsedDevice = field(default_factory=ConstantStepDevice)
    """Parameter that modify the behavior of the pulsed device."""

    forward: IOParameters = field(default_factory=IOParameters)
    """Input-output parameter setting for the forward direction."""

    backward: IOParameters = field(default_factory=IOParameters)
    """Input-output parameter setting for the backward direction."""

    update: UpdateParameters = field(default_factory=UpdateParameters)
    """Parameter for the update behavior."""

    def as_bindings(self) -> devices.AnalogTileParameter:
        """Return a representation of this instance as a simulator bindings object."""
        return tile_parameters_to_bindings(self)


@dataclass
class UnitCellRPUConfig(MapableRPU, PrePostProcessingRPU, _PrintableMixin):
    """Configuration for an analog (unit cell) resistive processing unit."""

    tile_class: ClassVar[Type] = AnalogTile
    """Tile class that correspond to this RPUConfig."""

    bindings_class: ClassVar[Type] = devices.AnalogTileParameter

    device: UnitCell = field(default_factory=UnitCell)
    """Parameter that modify the behavior of the pulsed device."""

    forward: IOParameters = field(default_factory=IOParameters)
    """Input-output parameter setting for the forward direction."""

    backward: IOParameters = field(default_factory=IOParameters)
    """Input-output parameter setting for the backward direction."""

    update: UpdateParameters = field(default_factory=UpdateParameters)
    """Parameter for the parallel analog update behavior."""

    def as_bindings(self) -> devices.AnalogTileParameter:
        """Return a representation of this instance as a simulator bindings object."""
        return tile_parameters_to_bindings(self)


@dataclass
class InferenceRPUConfig(MapableRPU, PrePostProcessingRPU, _PrintableMixin):
    """Configuration for an analog tile that is used only for inference.

    Training is done in *hardware-aware* manner, thus using only the
    non-idealities of the forward-pass, but backward and update passes
    are ideal.

    During inference, statistical models of programming, drift
    and read noise can be used.
    """
    # pylint: disable=too-many-instance-attributes

    tile_class: ClassVar[Type] = InferenceTile
    """Tile class that correspond to this RPUConfig."""

    bindings_class: ClassVar[Type] = devices.AnalogTileParameter

    forward: IOParameters = field(default_factory=IOParameters)
    """Input-output parameter setting for the forward direction.

    This parameters govern the hardware definitions specifying analog
    MVM non-idealities.

    Note:

        This forward pass is applied equally in training and
        inference. In addition, materials effects such as drift and
        programming noise can be enabled during inference by
        specifying the ``noise_model``

    """

    noise_model: BaseNoiseModel = field(default_factory=PCMLikeNoiseModel)
    """Statistical noise model to be used during (realistic) inference.

    This noise models establishes a phenomenological model of the
    material which is applied to the weights during inference only, when
    ``program_analog_weights`` or ``drift_analog_weights`` is called.

    """

    drift_compensation: Optional[BaseDriftCompensation] = field(
        default_factory=GlobalDriftCompensation)
    """For compensating the drift during inference only."""

    clip: WeightClipParameter = field(default_factory=WeightClipParameter)
    """Parameter for weight clip.

    If a clipping type is set, the weights are clipped according to
    the type specified.

    Caution:

        The clipping type is set to ``None`` by default, setting
        parameters of the clipping will not be taken into account, if
        the clipping type is not specified.
    """

    remap: WeightRemapParameter = field(default_factory=WeightRemapParameter)
    """Parameter for remapping.

    Remapping can be enabled by specifying a remap ``type``. If
    enabled, it ensures that the weights are mapped maximally into the
    conductance units during training. It will be called after each mini-batch.
    """

    modifier: WeightModifierParameter = field(default_factory=WeightModifierParameter)

    """Parameter for weight modifier.

    If a modifier type is set, it is called once per mini-match in the
    ``post_update_step`` and modifies the weight in forward and
    backward direction for the next mini-batch during training, but
    updates hidden reference weights. In eval mode, the reference
    weights are used instead for forward.

    The modifier is used to do hardware-aware training, so that the
    model becomes more noise robust during inference (e.g. when the
    ``noise_model`` is employed).
    """

    # The following fields are not included in `__init__`, and should be
    # treated as read-only.

    device: IdealDevice = field(default_factory=IdealDevice,
                                init=False)
    """Parameter that modify the behavior of the pulsed device: ideal device."""

    backward: IOParameters = field(
        default_factory=lambda: IOParameters(is_perfect=True),
        init=False
    )
    """Input-output parameter setting for the backward direction: perfect."""

    update: UpdateParameters = field(
        default_factory=lambda: UpdateParameters(pulse_type=PulseType.NONE),
        init=False
    )
    """Parameter for the update behavior: ``NONE`` pulse type."""

    def as_bindings(self) -> devices.AnalogTileParameter:
        """Return a representation of this instance as a simulator bindings object."""
        return tile_parameters_to_bindings(self)


@dataclass
class DigitalRankUpdateRPUConfig(MapableRPU, PrePostProcessingRPU, _PrintableMixin):
    """Configuration for an analog (unit cell) resistive processing unit
    where the rank update is done in digital.

    Note that for forward and backward, an analog crossbar is still
    used, and during update the digitally computed rank update is
    transferred to the analog crossbar using pulses.
    """

    tile_class: ClassVar[Type] = AnalogTile
    """Tile class that correspond to this RPUConfig."""

    bindings_class: ClassVar[Type] = devices.AnalogTileParameter

    device: DigitalRankUpdateCell = field(default_factory=DigitalRankUpdateCell)
    """Parameter that modify the behavior of the pulsed device."""

    forward: IOParameters = field(default_factory=IOParameters)
    """Input-output parameter setting for the forward direction."""

    backward: IOParameters = field(default_factory=IOParameters)
    """Input-output parameter setting for the backward direction."""

    update: UpdateParameters = field(default_factory=UpdateParameters)
    """Parameter for the analog part of the update, that is the transfer
    from the digital buffer to the devices."""

    def as_bindings(self) -> devices.AnalogTileParameter:
        """Return a representation of this instance as a simulator bindings object."""
        return tile_parameters_to_bindings(self)



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Some more tests for specific tiles."""

from torch import ones, Tensor
from torch.nn.functional import mse_loss

from aihwkit.simulator.configs.devices import SoftBoundsDevice
from aihwkit.simulator.configs.compounds import (
    TransferCompound, ReferenceUnitCell,
)

from aihwkit.simulator.configs.configs import UnitCellRPUConfig
from aihwkit.optim import AnalogSGD

from .helpers.decorators import parametrize_over_layers
from .helpers.layers import Linear, LinearCuda
from .helpers.testcases import ParametrizedTestCase
from .helpers.tiles import FloatingPoint


@parametrize_over_layers(
    layers=[Linear, LinearCuda],
    tiles=[FloatingPoint],
    biases=['analog', 'digital', None]
)
class TransferCompoundTest(ParametrizedTestCase):
    """Tests for transfer compound."""

    @staticmethod
    def get_transfer_compound(gamma, **kwargs):
        """Get a Tiki-taka compound with reference cell """
        def custom_device(**kwargs):
            """Custom device """
            return SoftBoundsDevice(w_max_dtod=0.0, w_min_dtod=0.0, w_max=1.0, w_min=-1.0, **kwargs)

        rpu_config = UnitCellRPUConfig(
            device=TransferCompound(

                # Devices that compose the Tiki-taka compound.
                unit_cell_devices=[
                    # fast "A" matrix
                    ReferenceUnitCell([custom_device(**kwargs), custom_device(**kwargs)]),
                    # slow "C" matrix
                    ReferenceUnitCell([custom_device(**kwargs), custom_device(**kwargs)])
                ],
                gamma=gamma,
            )
        )
        return rpu_config

    def test_hidden_parameter_setting(self):
        """Test hidden parameter set."""
        # pylint: disable=invalid-name

        for gamma in [0.0, 0.1]:
            model = self.get_layer(rpu_config=self.get_transfer_compound(gamma))

            weight, bias = model.get_weights()
            model.set_weights(weight * 0.0, bias * 0.0 if bias is not None else None)

            params = model.analog_tile.get_hidden_parameters()
            shape = params['hidden_weights_0_0'].shape

            # just dummy settings
            a, b, c, d = 0.47, 0.21, 0.64, 0.12
            params['hidden_weights_0_0'] = a*ones(*shape)  # A
            params['hidden_weights_1_0'] = b*ones(*shape)  # A ref
            params['hidden_weights_0_1'] = c*ones(*shape)  # C
            params['hidden_weights_1_1'] = d*ones(*shape)  # C_ref

            model.analog_tile.set_hidden_parameters(params)

            weight, bias = model.get_weights()

            # should be
            if self.digital_bias:
                self.assertEqual(bias[0], 0.0)
            if self.bias and not self.digital_bias:
                self.assertEqual(bias[0], gamma*(a - b) + c - d)

            self.assertEqual(weight[0][0], gamma*(a - b) + c - d)

    def test_decay(self):
        """Test hidden parameter set."""
        # pylint: disable=invalid-name, too-many-locals

        lifetime = 100.  # initial setting (needs to be larger 1)
        gamma = 0.1
        reset_bias = 0.1  # decay shift
        rpu_config = self.get_transfer_compound(gamma=gamma,
                                                lifetime=lifetime,
                                                lifetime_dtod=0.0,
                                                reset=reset_bias)

        model = self.get_layer(in_features=2, out_features=1, rpu_config=rpu_config)

        weight, bias = model.get_weights()
        model.set_weights(weight * 0.0, bias * 0.0 if bias is not None else None)

        params = model.analog_tile.get_hidden_parameters()
        shape = params['hidden_weights_0_0'].shape

        # just dummy settings
        a, b, c, d = 0.47, 0.21, 0.64, 0.12
        params['hidden_weights_0_0'] = a * ones(*shape)  # A
        params['hidden_weights_1_0'] = b * ones(*shape)  # A ref
        params['hidden_weights_0_1'] = c * ones(*shape)  # C
        params['hidden_weights_1_1'] = d * ones(*shape)  # C_ref

        # explicitly set the decay scales (which is 1-1/lifetime)
        a_dcy, b_dcy, c_dcy, d_dcy = 0.95, 0.78, 0.93, 0.92
        params['decay_scales_0_0'] = a_dcy * ones(*shape)  # A
        params['decay_scales_1_0'] = b_dcy * ones(*shape)  # A ref
        params['decay_scales_0_1'] = c_dcy * ones(*shape)  # C
        params['decay_scales_1_1'] = d_dcy * ones(*shape)  # C_ref

        model.analog_tile.set_hidden_parameters(params)

        # LR set to zero. Only lifetime will be applied
        opt = AnalogSGD(model.parameters(), lr=0.0)

        x_b = Tensor([[0.1, 0.2], [0.2, 0.4]])
        y_b = Tensor([[0.3], [0.6]])

        if self.use_cuda:
            model = model.cuda()
            x_b = x_b.cuda()
            y_b = y_b.cuda()

        epochs = 2
        for _ in range(epochs):
            opt.zero_grad()
            pred = model(x_b)
            loss = mse_loss(pred, y_b)

            loss.backward()
            opt.step()

        weight, bias = model.get_weights()

        # reference values
        a = (a - reset_bias) * pow(a_dcy, epochs) + reset_bias
        b = (b - reset_bias) * pow(b_dcy, epochs) + reset_bias
        c = (c - reset_bias) * pow(c_dcy, epochs) + reset_bias
        d = (d - reset_bias) * pow(d_dcy, epochs) + reset_bias

        if self.digital_bias:
            self.assertAlmostEqual(bias[0].item(), 0.0)
        if self.bias and not self.digital_bias:
            self.assertAlmostEqual(bias[0].item(), gamma*(a - b) + c - d, 5)

        self.assertAlmostEqual(weight[0][0].item(), gamma*(a - b) + c - d, 5)



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Noise model info in rpu_config to neural network model"""

# pylint: disable=no-name-in-module,import-error
from aihwkit.cloud.converter.definitions.i_input_file_pb2 import (  # type: ignore[attr-defined]
    NoiseModelProto
)


# pylint: disable=too-many-instance-attributes
class NoiseModelInfo:
    """Data only class for fields from protobuf NoiseModelProto message"""

    PCM = 'pcm'
    GENERIC = 'generic'

    def __init__(self, nm_proto: NoiseModelProto):  # type: ignore[valid-type]
        """Constructor for this class"""

        type_ = nm_proto.WhichOneof('item')  # type: ignore[attr-defined]

        info = None
        if type_ == 'pcm':
            # pcm does NOT have 2 extra fields
            info = nm_proto.pcm  # type: ignore[attr-defined]
        else:
            # generic HAS 2 extra fields
            info = nm_proto.generic  # type: ignore[attr-defined]

        self.device_id = info.device_id
        self.programming_noise_scale = info.programming_noise_scale
        self.read_noise_scale = info.read_noise_scale
        self.drift_scale = info.drift_scale
        self.drift_compensation = info.drift_compensation
        self.poly_first_order_coef = info.poly_first_order_coef
        self.poly_second_order_coef = info.poly_second_order_coef
        self.poly_constant_coef = info.poly_constant_coef

        self._drift_mean = -1.1
        self._drift_std = -1.1

        # Generic device has two extra field.
        if info.device_id == self.GENERIC:
            self._drift_mean = info.drift_mean
            self._drift_std = info.drift_std

    def _assert_generic(self) -> None:
        """Check is device is generic"""

        assert self.device_id == self.GENERIC, \
            "device_id does not have value '{}'".format(self.GENERIC)

    @property
    def drift_mean(self) -> float:
        """Enforce access to drift_mean if this is a generic device"""

        self._assert_generic()
        return self._drift_mean

    @property
    def drift_std(self) -> float:
        """Enforce access to drift_std if this is a generic device"""

        self._assert_generic()
        return self._drift_std
# pylint: enable=too-many-instance-attributes



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Base conductance converter for the phenomenological noise models for inference."""

from typing import Dict, List, Tuple

from torch import Tensor
from torch.autograd import no_grad


class BaseConductanceConverter:
    """Base class for converting DNN weights into conductances."""

    @no_grad()
    def convert_to_conductances(self, weights: Tensor) -> Tuple[List[Tensor], Dict]:
        """Convert a weight matrix into conductances.

        Caution:
            The conversion is assumed deterministic and repeatable.

        Args:
            weights: weight matrix tensor.

        Returns:
            Tuple of the list of conductance tensors and a params
            dictionary that is used for the reverse conversion.
        """
        raise NotImplementedError

    @no_grad()
    def convert_back_to_weights(self, conductances: List[Tensor], params: Dict) -> Tensor:
        """Convert a matrix of conductances into weights.

        Caution:
            The conversion is assumed deterministic and repeatable.

        Args:
            conductances: list of conductance tensors representing a weight matrix
            params: param dictionary that was returned from the conversion

        Returns:
            weight matrix
        """
        raise NotImplementedError

    def __eq__(self, other: object) -> bool:
        return self.__class__ == other.__class__ and self.__dict__ == other.__dict__



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=too-many-locals, too-many-branches

"""Fitting utilities.

This module includes fitting utilities for ``aihwkit``.
Using this module has extra dependencies that can be installed via the
extras mechanism::

    pip install aihwkit[fitting]
"""

from typing import Union, Dict, TypeVar, Tuple, Optional, List, Any
from copy import deepcopy
from dataclasses import fields

from numpy import array, concatenate, newaxis, ndarray
from torch import from_numpy, ones, stack, float32
from lmfit import minimize, Parameters, report_fit

from aihwkit.exceptions import ArgumentError, ConfigError
from aihwkit.simulator.configs.devices import PulsedDevice
from aihwkit.simulator.configs.configs import SingleRPUConfig
from aihwkit.simulator.tiles.analog import AnalogTile

RPUConfigGeneric = TypeVar('RPUConfigGeneric')


def _apply_parameters_to_config(device_config: Union[PulsedDevice, RPUConfigGeneric],
                                params: Parameters) -> None:
    """Apply the fit parameters to the device config.

    Args:
         device_config: device config to be set (in place)
         params: lmfit.Parameters structure

    Raises: ConfigError if parameter was not found"""

    parvals = params.valuesdict()
    if isinstance(device_config, PulsedDevice):
        device = device_config
    else:
        device = getattr(device_config, 'device')  # type ignore

    for par, value in parvals.items():
        if not hasattr(device, par):
            raise ConfigError(f" Cannot find parameter '{par}' in device config.")
        setattr(device, par, value)


def fit_measurements(parameters: Union[Dict, Parameters],
                     pulse_data: Union[Tuple[ndarray], ndarray],
                     response_data: Union[Tuple[ndarray], ndarray],
                     device_config: Union[PulsedDevice, RPUConfigGeneric],
                     suppress_device_noise: bool = True,
                     max_pulses: Optional[int] = 1,
                     n_traces: int = 1,
                     fit_weights: Optional[Union[Tuple[int], int]] = None,
                     method: str = 'powell',
                     verbose: bool = False,
                     **fit_kwargs: Any
                     ) -> Tuple[Any,
                                Union[PulsedDevice, RPUConfigGeneric],
                                List[ndarray]]:
    """Fit pulse response measurement to the given device model using lmfit.

    For example:

    .. code-block:: python

        # responses are conductance data in response to pulses (-1, 1)

        # choose device model and parameter to fit
        device_config = SoftBoundsDevice(w_min=-1.0, w_max=1.0)
        params = {'dw_min': (0.1, 0.001, 5.0),
                  'up_down': (0.0, -0.99, 0.99),
                  'w_max': (1.0, 0.1, 5.0)}

        # fit the response
        fit_res, fit_device_config, model_response = fit_measurements(
            params, pulses, responses,
            device_config=device_config,
            suppress_device_noise=True,
            method='powell',
            fit_weights=fit_weights,
        )
        # fit parameter
        print(fit_res.params.valuesdict())
        # device of best fit
        print(fit_device_config)


    Args:

        parameters: Parameter to vary. Dictionary with parameter names
            (attributes of the device config). Each value is either a
            single value (thus only set, not varied) or a tuple
            ``(x_init, x_min, x_max)``. ``lmfit.Parameters`` class can
            also given directly.

        pulse_data: Pulse data, ie array of number of pulses in up
            (pos) or down (neg) direction. Can be a tuple of multiple
            measurements

        response_data: Corresponfing measured responses to the pulses
            given by ``pulse_data`` as numpy array or list.

            Caution:
                ``axes=1`` can be used for multiple device
                fit. However, then all pulse data needs to have the
                same axis=0 dimension

        device_config: base device configuration

        suppress_device_noise: sets all dtod and std parameters of the device to 0

        n_traces: how many traces to simulate simulaenously

        max_pulses: constrain the number of pulses given.

        fit_weights: the weightening of the individual response traces
            in the loss function

        method: fitting method from ``lmfit`` (default "powell")

        verbose: whether to print fitting results

        fit_kwargs: additional parameter passed to ``lmfit.minimize``

    Returns:

         fit_results: Result of the fit in ``lmfit`` format
         device_config: Device config with found parameter applied
         model_response: Model response of parameter fit

    Raises:
         ArgumentError: in case wrong arguments are given

    """

    if isinstance(pulse_data, tuple) != isinstance(response_data, tuple):
        raise ArgumentError("Either all data inputs need to be tuples or None. ")

    device_config = deepcopy(device_config)
    if isinstance(device_config, PulsedDevice):
        rpu_config = SingleRPUConfig(device=device_config)

        # single pulse mode
        if max_pulses is not None:
            rpu_config.update.desired_bl = max_pulses
            rpu_config.update.update_bl_management = False
            rpu_config.update.update_management = False

    else:
        rpu_config = device_config  # type: ignore

    if suppress_device_noise:
        for field in fields(rpu_config.device):
            if field.name.endswith('dtod') or field.name.endswith('std'):
                setattr(rpu_config.device, field.name, 0.0)

    params = Parameters()
    if isinstance(parameters, Parameters):
        params = parameters
    elif isinstance(parameters, dict):
        for par, values in parameters.items():
            if isinstance(values, tuple):
                x_init, x_min, x_max = values
                params.add(par, value=x_init, min=x_min, max=x_max, vary=True)
            else:
                params.add(par, value=values, vary=False)
    else:
        raise ArgumentError("Expect dict or Parameters for parmeters.")

    # fit parameters
    args = (pulse_data, response_data, rpu_config, n_traces, fit_weights, verbose)
    result = minimize(model_response, params, args=args,
                      method=method, **fit_kwargs)
    if verbose:
        report_fit(result)

    best_model_res = model_response(result.params, *args, only_response=True)

    _apply_parameters_to_config(device_config, result.params)
    return result, device_config, best_model_res  # type: ignore


def model_response(params: Parameters,
                   pulse_data: Union[Tuple[ndarray], ndarray],
                   response_data: Union[Tuple[ndarray], ndarray],
                   rpu_config: RPUConfigGeneric,
                   n_traces: int = 1,
                   fit_weights: Optional[Union[Tuple[int], int]] = None,
                   verbose: bool = True,
                   only_response: bool = False,
                   ) -> Union[ndarray, List[ndarray]]:
    """Compute the model respunses given the pulses.

    Args:
        params: ``lmfit.Parameters`` of the current parameter setting

        pulse_data: Pulse data, ie array of number of pulses in up
            (pos) or down (neg) direction. Can be a tuple of multiple
            measurements

        response_data: Corresponfing measured responses to the pulses
            given by ``pulse_data`` as numpy array or list.

            Caution:
                ``axes=1`` can be used for multiple device
                fit. However, then all pulse data needs to have the
                same axis=0 dimension

        rpu_config: base device configuration (will be modified)

        fit_weights: the weightening of the individual response traces
            in the loss function

        n_traces: how many traces to simulate simulaenously

        verbose: whether to print std of deviation

        only_response: whether to returns a list of model response
           instead of the deviation

    Returns:
        deviation vector or list of model responses (weight traces)

    Note:
        overwrites the given rpu_config

    """

    _apply_parameters_to_config(rpu_config, params)

    # likley somewhat inefficient since we need to always create a new
    # tile, repeats are quick though
    no_list = False
    if not isinstance(pulse_data, tuple):
        pulse_data = (pulse_data, )
        response_data = (response_data, )  # type: ignore
        if fit_weights is not None:
            fit_weights = (fit_weights, )  # type: ignore
        no_list = True

    numpy_pulses = array(pulse_data[0])
    n_devices = 1
    if numpy_pulses.ndim > 1:
        n_devices = numpy_pulses.shape[1]

    analog_tile = AnalogTile(n_traces, n_devices, rpu_config)  # type: ignore
    analog_tile.set_learning_rate(1)

    deviation = array([])
    model_responses = []
    for idx, (numpy_pulses, response) in enumerate(zip(pulse_data, response_data)):
        w_init = response[0, :]
        weights = from_numpy(array(w_init).flatten()[newaxis, :]
                             ).to(dtype=float32) * ones((n_traces, n_devices), dtype=float32)
        analog_tile.set_weights(weights)
        pulses = from_numpy(numpy_pulses).to(dtype=float32)
        w_trace = [weights]
        for pulse in pulses[:-1]:
            analog_tile.update(pulse * ones(n_devices, dtype=float32),
                               -ones((n_traces), dtype=float32))
            w_trace.append(analog_tile.tile.get_weights())

        stacked_w_trace = stack(w_trace).cpu().numpy()

        # compute square error
        num_samples = response.shape[0]
        avg_w_trace = stacked_w_trace.mean(axis=1)[:num_samples, :]
        model_responses.append(avg_w_trace)
        dev = (avg_w_trace - response)
        if fit_weights is not None:
            dev = dev * array(fit_weights[idx])[:dev.shape[1]]  # type: ignore
        deviation = concatenate([deviation, dev.flatten()])

    if only_response:
        if no_list:
            return model_responses[0]
        return model_responses
    if verbose:
        print(deviation.std())
    return deviation



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Helper class for adding rpu_config to neural network model"""
# pylint: disable=no-name-in-module,import-error

from aihwkit.cloud.converter.definitions.i_input_file_pb2 import (  # type: ignore[attr-defined]
    AnalogProto
)


# pylint: disable=too-few-public-methods)
class AnalogInfo:
    """Data class for fields from protobuf AnalogProto message"""

    def __init__(self, a_info: AnalogProto):  # type: ignore[valid-type]
        """Constructor for this class"""

        # all three fields are required
        self.output_noise_strength = a_info.output_noise_strength  # type: ignore[attr-defined]
        self.adc = a_info.adc  # type: ignore[attr-defined]
        self.dac = a_info.dac  # type: ignore[attr-defined]



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=too-many-instance-attributes

"""Phenomenological noise models for PCM devices for inference."""

from copy import deepcopy
from typing import List, Optional

from numpy import log as numpy_log
from numpy import sqrt
from torch import abs as torch_abs
from torch import clamp, log, randn_like, Tensor
from torch.autograd import no_grad

from aihwkit.inference.noise.base import BaseNoiseModel
from aihwkit.inference.converter.base import BaseConductanceConverter
from aihwkit.inference.converter.conductance import SinglePairConductanceConverter

_ZERO_CLIP = 1e-7


class PCMLikeNoiseModel(BaseNoiseModel):
    r"""Noise model that was fitted and characterized on real PCM devices.

    Expected weight noise at assumed time of inference with expected
    programming noise at 0.

    The statistical noise model is based on measured PCM devices.

    Args:
        prog_coeff: programming polynomial coeffs in :math:`\mu S`, c(0) + c(1)*gt + c(2)*gt^2)
        g_converter: instantiated class of the conductance converter (defaults to single pair)
        g_max: In :math:`\mu S`, the maximal conductance, ie the value
            the absolute max of the weights will be mapped to.
        t_read: parameter of the 1/f fit (in seconds)
        t_0: parameter of the drift fit (first reading time)

            Note:
                The ``t_inference`` is relative to this time `t0`
                e.g. t_inference counts from the completion of the programming
                of a device.
        prog_noise_scale: scale for the programming noise
        read_noise_scale: scale for the read and accumulated noise
        drift_scale: scale for the  drift coefficient

    """

    def __init__(
            self,
            prog_coeff: Optional[List[float]] = None,
            g_converter: Optional[BaseConductanceConverter] = None,
            g_max: Optional[float] = None,
            t_read: float = 250.0e-9,
            t_0: float = 20.0,
            prog_noise_scale: float = 1.0,
            read_noise_scale: float = 1.0,
            drift_scale: float = 1.0,
    ):
        g_converter = deepcopy(g_converter) or SinglePairConductanceConverter(g_max=g_max)
        super().__init__(g_converter)

        self.g_max = getattr(self.g_converter, 'g_max', g_max)

        if self.g_max is None:
            raise ValueError('g_max cannot be established from g_converter')

        self.prog_coeff = [0.26348, 1.9650, -1.1731] if prog_coeff is None else prog_coeff
        self.t_0 = t_0
        self.t_read = t_read
        self.prog_noise_scale = prog_noise_scale
        self.read_noise_scale = read_noise_scale
        self.drift_scale = drift_scale

    def __str__(self) -> str:
        return ('{}(prog_coeff={}, g_converter={}, g_max={:1.2f}, t_read={}, '
                't_0={:1.2f}, prog_noise_scale={}, '
                'read_noise_scale={}, drift_scale={})').format(  # type: ignore
                    self.__class__.__name__, self.prog_coeff, self.g_converter,
                    self.g_max,
                    self.t_read, self.t_0, self.prog_noise_scale,
                    self.read_noise_scale, self.drift_scale)

    @no_grad()
    def apply_programming_noise_to_conductance(self, g_target: Tensor) -> Tensor:
        """Apply programming noise to a target conductance Tensor.

        Programming noise with additive Gaussian noise with
        conductance dependency of the variance given by a 2-degree
        polynomial.
        """
        mat = 1
        sig_prog = self.prog_coeff[0]
        for coeff in self.prog_coeff[1:]:
            mat *= g_target / self.g_max
            sig_prog += mat * coeff

        g_prog = g_target + self.prog_noise_scale * sig_prog * randn_like(g_target)
        g_prog.clamp_(min=0.0)  # no negative conductances allowed

        return g_prog

    @no_grad()
    def generate_drift_coefficients(self, g_target: Tensor) -> Tensor:
        """Return drift coefficients ``nu`` based on PCM measurements."""
        g_relative = clamp(torch_abs(g_target / self.g_max), min=_ZERO_CLIP)

        # gt should be normalized wrt g_max
        mu_drift = (-0.0155 * log(g_relative) + 0.0244).clamp(min=0.049, max=0.1)
        sig_drift = (-0.0125 * log(g_relative) - 0.0059).clamp(min=0.008, max=0.045)
        nu_drift = torch_abs(mu_drift + sig_drift * randn_like(g_relative)).clamp(min=0.0)

        return nu_drift * self.drift_scale

    @no_grad()
    def apply_drift_noise_to_conductance(
            self,
            g_prog: Tensor,
            nu_drift: Tensor,
            t_inference: float
    ) -> Tensor:
        """Apply the noise and drift up to the assumed inference time
        point based on PCM measurements."""
        t = t_inference + self.t_0

        # drift
        if t > self.t_0:
            g_drift = g_prog * ((t / self.t_0) ** (- nu_drift))
        else:
            g_drift = g_prog

        # expected accumulated 1/f noise since start of programming at t=0
        if t > 0:
            q_s = (0.0088 / ((torch_abs(g_prog) /
                              self.g_max) ** 0.65).clamp(min=1e-3)).clamp(max=0.2)
            sig_noise = q_s * sqrt(numpy_log((t + self.t_read) / (2 * self.t_read)))
            g_final = g_drift + torch_abs(g_drift) * self.read_noise_scale \
                * sig_noise * randn_like(g_prog)
        else:
            g_final = g_prog

        return g_final.clamp(min=0.0)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: common.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import onnx_common_pb2 as onnx__common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0c\x63ommon.proto\x12\x05\x61ihwx\x1a\x11onnx_common.proto\"F\n\x0ePrimitiveProto\x12\n\n\x02id\x18\x01 \x02(\t\x12(\n\targuments\x18\x02 \x03(\x0b\x32\x15.aihwx.AttributeProto\"F\n\x0eTransformProto\x12\n\n\x02id\x18\x01 \x02(\t\x12(\n\targuments\x18\x02 \x03(\x0b\x32\x15.aihwx.AttributeProto\"O\n\x17\x41\x63tivationFunctionProto\x12\n\n\x02id\x18\x01 \x02(\t\x12(\n\targuments\x18\x02 \x03(\x0b\x32\x15.aihwx.AttributeProto\"m\n\nLayerProto\x12\n\n\x02id\x18\x01 \x02(\t\x12(\n\targuments\x18\x02 \x03(\x0b\x32\x15.aihwx.AttributeProto\x12)\n\nstate_dict\x18\x03 \x03(\x0b\x32\x15.aihwx.AttributeProto\"\x86\x01\n\x19LayerOrActivationFunction\x12=\n\x13\x61\x63tivation_function\x18\x01 \x01(\x0b\x32\x1e.aihwx.ActivationFunctionProtoH\x00\x12\"\n\x05layer\x18\x02 \x01(\x0b\x32\x11.aihwx.LayerProtoH\x00\x42\x06\n\x04item\"F\n\x0eOptimizerProto\x12\n\n\x02id\x18\x01 \x02(\t\x12(\n\targuments\x18\x02 \x03(\x0b\x32\x15.aihwx.AttributeProto\"F\n\x0eSchedulerProto\x12\n\n\x02id\x18\x01 \x02(\t\x12(\n\targuments\x18\x02 \x03(\x0b\x32\x15.aihwx.AttributeProto\"I\n\x11LossFunctionProto\x12\n\n\x02id\x18\x01 \x02(\t\x12(\n\targuments\x18\x02 \x03(\x0b\x32\x15.aihwx.AttributeProto\";\n\x07Network\x12\x30\n\x06layers\x18\x01 \x03(\x0b\x32 .aihwx.LayerOrActivationFunction\"(\n\x07Version\x12\x0e\n\x06schema\x18\x01 \x02(\x03\x12\r\n\x05opset\x18\x02 \x02(\x03')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'common_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _PRIMITIVEPROTO._serialized_start=42
  _PRIMITIVEPROTO._serialized_end=112
  _TRANSFORMPROTO._serialized_start=114
  _TRANSFORMPROTO._serialized_end=184
  _ACTIVATIONFUNCTIONPROTO._serialized_start=186
  _ACTIVATIONFUNCTIONPROTO._serialized_end=265
  _LAYERPROTO._serialized_start=267
  _LAYERPROTO._serialized_end=376
  _LAYERORACTIVATIONFUNCTION._serialized_start=379
  _LAYERORACTIVATIONFUNCTION._serialized_end=513
  _OPTIMIZERPROTO._serialized_start=515
  _OPTIMIZERPROTO._serialized_end=585
  _SCHEDULERPROTO._serialized_start=587
  _SCHEDULERPROTO._serialized_end=657
  _LOSSFUNCTIONPROTO._serialized_start=659
  _LOSSFUNCTIONPROTO._serialized_end=732
  _NETWORK._serialized_start=734
  _NETWORK._serialized_end=793
  _VERSION._serialized_start=795
  _VERSION._serialized_end=835
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: i_input_file.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import i_common_pb2 as i__common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12i_input_file.proto\x12\x11\x61ihwx.inferencing\x1a\x0ei_common.proto\"1\n\x07\x44\x61taset\x12\x12\n\ndataset_id\x18\x01 \x02(\t\x12\x12\n\nbatch_size\x18\x02 \x01(\x03\"F\n\x0b\x41nalogProto\x12\x1d\n\x15output_noise_strength\x18\x01 \x02(\x02\x12\x0b\n\x03\x61\x64\x63\x18\x02 \x01(\x03\x12\x0b\n\x03\x64\x61\x63\x18\x03 \x01(\x03\"\xe9\x01\n\x08PCMProto\x12\x16\n\tdevice_id\x18\x01 \x02(\t:\x03pcm\x12\x1f\n\x17programming_noise_scale\x18\x02 \x02(\x02\x12\x18\n\x10read_noise_scale\x18\x03 \x02(\x02\x12\x13\n\x0b\x64rift_scale\x18\x04 \x02(\x02\x12\x1a\n\x12\x64rift_compensation\x18\x05 \x02(\x08\x12\x1d\n\x15poly_first_order_coef\x18\x06 \x02(\x02\x12\x1e\n\x16poly_second_order_coef\x18\x07 \x02(\x02\x12\x1a\n\x12poly_constant_coef\x18\x08 \x02(\x02\"\x98\x02\n\x0cGenericProto\x12\x1a\n\tdevice_id\x18\x01 \x02(\t:\x07generic\x12\x1f\n\x17programming_noise_scale\x18\x02 \x02(\x02\x12\x18\n\x10read_noise_scale\x18\x03 \x02(\x02\x12\x13\n\x0b\x64rift_scale\x18\x04 \x02(\x02\x12\x1a\n\x12\x64rift_compensation\x18\x05 \x02(\x08\x12\x1d\n\x15poly_first_order_coef\x18\x06 \x02(\x02\x12\x1e\n\x16poly_second_order_coef\x18\x07 \x02(\x02\x12\x1a\n\x12poly_constant_coef\x18\x08 \x02(\x02\x12\x12\n\ndrift_mean\x18\t \x02(\x02\x12\x11\n\tdrift_std\x18\n \x02(\x02\"y\n\x0fNoiseModelProto\x12*\n\x03pcm\x18\x01 \x01(\x0b\x32\x1b.aihwx.inferencing.PCMProtoH\x00\x12\x32\n\x07generic\x18\x02 \x01(\x0b\x32\x1f.aihwx.inferencing.GenericProtoH\x00\x42\x06\n\x04item\"\xb3\x01\n\x0bInferencing\x12\x19\n\x11inference_repeats\x18\x01 \x02(\x03\x12\x16\n\x0einference_time\x18\x02 \x02(\x03\x12\x33\n\x0b\x61nalog_info\x18\x03 \x02(\x0b\x32\x1e.aihwx.inferencing.AnalogProto\x12<\n\x10noise_model_info\x18\x04 \x02(\x0b\x32\".aihwx.inferencing.NoiseModelProto\"\xcc\x01\n\x0eInferenceInput\x12+\n\x07version\x18\x01 \x02(\x0b\x32\x1a.aihwx.inferencing.Version\x12+\n\x07\x64\x61taset\x18\x02 \x02(\x0b\x32\x1a.aihwx.inferencing.Dataset\x12+\n\x07network\x18\x03 \x02(\x0b\x32\x1a.aihwx.inferencing.Network\x12\x33\n\x0binferencing\x18\x04 \x02(\x0b\x32\x1e.aihwx.inferencing.Inferencing')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'i_input_file_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _DATASET._serialized_start=57
  _DATASET._serialized_end=106
  _ANALOGPROTO._serialized_start=108
  _ANALOGPROTO._serialized_end=178
  _PCMPROTO._serialized_start=181
  _PCMPROTO._serialized_end=414
  _GENERICPROTO._serialized_start=417
  _GENERICPROTO._serialized_end=697
  _NOISEMODELPROTO._serialized_start=699
  _NOISEMODELPROTO._serialized_end=820
  _INFERENCING._serialized_start=823
  _INFERENCING._serialized_end=1002
  _INFERENCEINPUT._serialized_start=1005
  _INFERENCEINPUT._serialized_end=1209
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-

# (C) Copyright 2020, 2021, 2022 IBM. All Rights Reserved.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Conductance converters for the phenomenological noise models for inference."""

from typing import Dict, List, Optional, Tuple

from torch import abs as torch_abs
from torch import Tensor
from torch.autograd import no_grad

from aihwkit.inference.converter.base import BaseConductanceConverter

_ZERO_CLIP = 1e-7


class SinglePairConductanceConverter(BaseConductanceConverter):
    r"""Single pair of devices.

    Assuming a single pair of devices per cross-point, taking positive
    and negative weights, respectively, where one device is always at
    0.

    Args:
        g_max: In :math:`\mu S`, the maximal conductance, ie the value
            the absolute max of the weights will be mapped to.
        g_min: In :math:`\mu S`, the minimal conductance, ie the value
            the logical zero of the weights will be mapped to.
    """

    def __init__(
            self,
            g_max: Optional[float] = None,
            g_min: Optional[float] = None
    ):
        self.g_max = 25.0 if g_max is None else g_max
        self.g_min = 0.0 if g_min is None else g_min
        self.scale_ratio = None

        if self.g_max < 0.:
            raise ValueError('g_max should be a positive value')
        if self.g_min < 0.:
            raise ValueError('g_min should be a positive value')
        if self.g_min > self.g_max:
            raise ValueError('g_min should be smaller than g_max')

    def __str__(self) -> str:
        return '{}(g_max={:1.2f}, g_min={:1.2f})'.format(
            self.__class__.__name__, self.g_max, self.g_min
        )

    @no_grad()
    def convert_to_conductances(self, weights: Tensor) -> Tuple[List[Tensor], Dict]:
        abs_max = torch_abs(weights).max()
        scale_ratio = (self.g_max - self.g_min) / abs_max.clamp(min=_ZERO_CLIP)
        scaled_weights = weights * scale_ratio

        conductances = [scaled_weights.clamp(min=0.0, max=self.g_max) + self.g_min,
                        (- scaled_weights).clamp(min=0.0, max=self.g_max) + self.g_min]
        params = {'scale_ratio': scale_ratio}

        return conductances, params

    @no_grad()
    def convert_back_to_weights(self, conductances: List[Tensor], params: Dict) -> Tensor:
        if len(conductances) != 2:
            raise ValueError('conductances must contain exactly two elements')
        if 'scale_ratio' not in params:
            raise ValueError('params do not contain scale_ratio')

        weights = ((conductances[0] - self.g_min) -
                   (conductances[1] - self.g_min)) / params['scale_ratio']

        return weights
